from typing import Union, Iterable

import numpy as np
import pandas as pn

from gempy_legacy.utils.meta import _setdoc_pro


class Faults(object):
    """
    Class that encapsulate faulting related content. Mainly, which surfaces/surfaces are faults. The fault network
    ---i.e. which faults offset other faults---and fault types---finite vs infinite.

    Args:
        series_fault(str, list[str]): Name of the series which are faults
        rel_matrix (numpy.array[bool]): 2D Boolean array with boolean logic. Rows affect (offset) columns

    Attributes:
       df (:class:`pn.core.frame.DataFrames`): Pandas data frame containing the series as index and if they are faults
        or not (otherwise they are lithologies) and in case of being fault if is finite
       faults_relations_df (:class:`pn.core.frame.DataFrames`): Pandas data frame containing the offsetting relations
        between each fault and the rest of the series (either other faults or lithologies)
       n_faults (int): Number of faults in the object
    """

    def __init__(self, series_fault=None, rel_matrix=None):

        self.df = pn.DataFrame(np.array([[False, False]]), index=pn.CategoricalIndex(['Default series']),
                               columns=['isFault', 'isFinite'], dtype=bool)

        self.faults_relations_df = pn.DataFrame(index=pn.CategoricalIndex(['Default series']),
                                                columns=pn.CategoricalIndex(['Default series', '']), dtype='bool')

        self.set_is_fault(series_fault=series_fault)
        self.set_fault_relation(rel_matrix=rel_matrix)
        self.n_faults = 0

        self._offset_faults = False

    def __repr__(self):
        return self.df.to_string()

    def _repr_html_(self):
        return self.df.to_html()

    def set_is_fault(self, series_fault: Union[str, list, np.ndarray] = None,
                     toggle=False, offset_faults=False):
        """
        Set a flag to the series that are faults.

        Args:
            series_fault(str, list[str]): Name of the series which are faults
            toggle (bool): if True, passing a name which is already True will set it False.
            offset_faults (bool): If True by default faults offset other faults

        Returns:
            :class:`gempy.core.data_modules.stack.Faults`

        """
        series_fault = np.atleast_1d(series_fault)
        self.df['isFault'].fillna(False, inplace=True)

        if series_fault is None:
            series_fault = self.count_faults(self.df.index)

        if series_fault[0] is not None:
            assert np.isin(series_fault, self.df.index).all(), 'series_faults must already ' \
                                                                                      'exist in the the series df.'
            if toggle is True:
                self.df.loc[series_fault, 'isFault'] = self.df.loc[series_fault, 'isFault'] ^ True
            else:
                self.df.loc[series_fault, 'isFault'] = True

            self.df['isFinite'] = np.bitwise_and(self.df['isFault'], self.df['isFinite'])

            self.set_default_faults_relations(offset_faults)
            # Update default fault relations
            for a_series in series_fault:
                col_pos = self.faults_relations_df.columns.get_loc(a_series)
                # set the faults offset all younger
                self.faults_relations_df.iloc[col_pos, col_pos + 1:] = True

                if offset_faults is False:
                    # set the faults does not offset the younger faults
                    self.faults_relations_df.iloc[col_pos] = ~self.df['isFault'] & \
                                                             self.faults_relations_df.iloc[col_pos]

        self.n_faults = self.df['isFault'].sum()

        return self

    def set_default_faults_relations(self, offset_faults:bool=None):
        if offset_faults is not None:
            self._offset_faults = offset_faults

        offset_faults = self._offset_faults

        try:
            # Update default fault relations
            for a_series in self.df.groupby('isFault').get_group(True).index:
                col_pos = self.faults_relations_df.columns.get_loc(a_series)
                # set the faults offset all younger
                self.faults_relations_df.iloc[col_pos, col_pos + 1:] = True

                if offset_faults is False:
                    # set the faults does not offset the younger faults
                    self.faults_relations_df.iloc[col_pos] = ~self.df['isFault'] & \
                                                             self.faults_relations_df.iloc[col_pos]
            return True

        except KeyError:
            return False

    def set_is_finite_fault(self, series_finite: Union[str, list, np.ndarray] = None, toggle=False):
        """
        Toggles given series' finite fault property.

        Args:
            series_finite (str, list[str]): Name of the series which are finite
            toggle (bool): if True, passing a name which is already True will set it False.

        Returns:
            :class:`gempy.core.data_modules.stack.Faults`
        """
        if series_finite[0] is not None:
            # check if given series is/are in dataframe
            assert np.isin(series_finite, self.df.index).all(), "series_fault must already exist" \
                                                                "in the series DataFrame."
            assert self.df.loc[series_finite].isFault.all(), "series_fault contains non-fault series" \
                                                             ", which can't be set as finite faults."
            # if so, toggle True/False for given series or list of series
            if toggle is True:
                self.df.loc[series_finite, 'isFinite'] = self.df.loc[series_finite, 'isFinite'] ^ True
            else:
                self.df.loc[series_finite, 'isFinite'] = True

        return self

    def set_fault_relation(self, rel_matrix=None):
        """Method to set the df that offset a given sequence and therefore also another fault.

        Args:
            rel_matrix (numpy.array[bool]): 2D Boolean array with boolean logic.
             Rows affect (offset) columns

        Returns:
            :class:`gempy.core.data_modules.stack.Faults.faults_relations_df`

        """
        # TODO: block the lower triangular matrix of being changed
        if rel_matrix is None:
            rel_matrix = np.zeros((self.df.index.shape[0],
                                   self.df.index.shape[0]))
        else:
            assert type(rel_matrix) is np.ndarray, 'rel_matrix muxt be a 2D numpy array'
        self.faults_relations_df = pn.DataFrame(rel_matrix, index=self.df.index,
                                                columns=self.df.index, dtype='bool')

        self.faults_relations_df.iloc[np.tril(np.ones(self.df.index.shape[0])).astype(bool)] = False

        return self.faults_relations_df

    @staticmethod
    def count_faults(list_of_names):
        """
        Read the string names of the surfaces to detect automatically the number of df if the name
        fault is on the name.
        """
        faults_series = []
        for i in list_of_names:
            try:
                if ('fault' in i or 'Fault' in i) and 'Default' not in i:
                    faults_series.append(i)
            except TypeError:
                pass
        return faults_series


@_setdoc_pro(Faults.__doc__)
class Series(object):
    """ Class that contains the functionality and attributes related to the series. Notice that series does not only
    refers to stratigraphic series but to any set of surfaces which will be interpolated together (comfortably).

    Args:
        faults (:class:`Faults`): [s0]
        series_names(Optional[list]): name of the series. They are also ordered

    Attributes:
        df (:class:`pn.core.frame.DataFrames`): Pandas data frame containing the series and the surfaces contained
            on them. TODO describe df columns
        faults (:class:`Faults`)
    """

    def __init__(self, faults, series_names: list = None):

        self.faults = faults

        if series_names is None:
            series_names = ['Default series']

        self.df = pn.DataFrame(np.array([[1, np.nan]]), index=pn.CategoricalIndex(series_names, ordered=False),
                               columns=['order_series', 'BottomRelation'])

        self.df['order_series'] = self.df['order_series'].astype(int)
        self.df['BottomRelation'] = pn.Categorical(['Erosion'], categories=['Erosion', 'Onlap', 'Fault'])
        self.df['isActive'] = False

    def __repr__(self):
        return self.df.to_string()

    def _repr_html_(self):
        return self.df.to_html()

    def reset_order_series(self):
        """
        Reset the column order series to monotonic ascendant values.
        """
        self.df['order_series'] = pn.RangeIndex(1, self.df.shape[0] + 1) # pandas 1.4 change
        #self.df.at[:, 'order_series'] = pn.RangeIndex(1, self.df.shape[0] + 1) # pandas 1.3 version

    @_setdoc_pro(reset_order_series.__doc__)
    def set_series_index(self, series_order: Union[list, np.ndarray], reset_order_series=True):
        """
        Rewrite the index of the series df

        Args:
            series_order (list): List with names and order of series.
            reset_order_series (bool): if true [s0]

        Returns:
             :class:`Series`: Series
        """
        if type(series_order) is list or type(series_order) is np.ndarray:
            list_of_series = np.atleast_1d(series_order)
        else:
            raise AttributeError('series_order is not neither list or SurfacePoints object.')

        series_idx = list_of_series

        # Categorical index does not have inplace
        # This update the categories
        self.df.index = self.df.index.set_categories(series_idx, rename=True)
        self.faults.df.index = self.faults.df.index.set_categories(series_idx, rename=True)
        self.faults.faults_relations_df.index = self.faults.faults_relations_df.index.set_categories(
            series_idx, rename=True)
        self.faults.faults_relations_df.columns = self.faults.faults_relations_df.columns.set_categories(
            series_idx, rename=True)

        # But we need to update the values too
        for c in series_order:
            try:
                self.df.loc[c] = [-1, 'Erosion', False, False, False]
            # This is in case someone use the old series
            except ValueError:
                self.df.loc[c] = [-1, 'Erosion', False]
                self.faults.df.loc[c, ['isFault', 'isFinite']] = [False, False]
            self.faults.faults_relations_df.loc[c, c] = False

        self.faults.faults_relations_df.fillna(False, inplace=True)

        if reset_order_series is True:
            self.reset_order_series()
        return self

    def set_bottom_relation(self, series_list: Union[str, list], bottom_relation: Union[str, list]):
        """Set the bottom relation between the series and the one below.

        Args:
            series_list (str, list): name or list of names of the series to apply the functionality
            bottom_relation (str{Onlap, Erode, Fault}, list[str]):

        Returns:
            :class:`gempy.core.data_modules.stack.Stack`
        """
        self.df.loc[series_list, 'BottomRelation'] = bottom_relation

        if self.faults.df.loc[series_list, 'isFault'] is True:
            self.faults.set_is_fault(series_list, toggle=True)

        elif bottom_relation == 'Fault':
            self.faults.df.loc[series_list, 'isFault'] = True
        return self

    @_setdoc_pro(reset_order_series.__doc__)
    def add_series(self, series_list: Union[str, list], reset_order_series=True):
        """ Add series to the df

        Args:
            series_list (str, list): name or list of names of the series to apply the functionality
            reset_order_series (bool): if true [s0]

        Returns:
            Series
        """
        series_list = np.atleast_1d(series_list)

        # Remove from the list categories that already exist
        series_list = series_list[~np.in1d(series_list, self.df.index.categories)]

        idx = self.df.index.add_categories(series_list)
        self.df.index = idx
        self.update_faults_index_rename()

        for c in series_list:
            # This is in case someone wants to run the old series
            try:
                self.df.loc[c] = [-1, 'Erosion', False, False, False]
            except ValueError:
                self.df.loc[c] = [-1, 'Erosion', False]
                self.faults.df.loc[c, ['isFault', 'isFinite']] = [False, False]
            self.faults.faults_relations_df.loc[c, c] = False

        self.faults.faults_relations_df.fillna(False, inplace=True)

        if reset_order_series is True:
            self.reset_order_series()
        return self

    @_setdoc_pro([reset_order_series.__doc__, pn.DataFrame.drop.__doc__])
    def delete_series(self, indices: Union[str, Iterable], reset_order_series=True):
        """[s1]

        Args:
            indices (str, list): name or list of names of the series to apply the functionality
            reset_order_series (bool): if true [s0]

        Returns:
            Series
        """
        self.df.drop(indices, inplace=True)
        # If we are using the Stack class it is just one element in memory
        try:
            self.faults.df.drop(indices, inplace=True)
        except KeyError:
            pass

        self.faults.faults_relations_df.drop(indices, axis=0, inplace=True)
        self.faults.faults_relations_df.drop(indices, axis=1, inplace=True)

        idx = self.df.index.remove_unused_categories()
        self.df.index = idx
        self.update_faults_index_rename()

        if reset_order_series is True:
            self.reset_order_series()
        return self

    @_setdoc_pro(pn.CategoricalIndex.rename_categories.__doc__)
    def rename_series(self, new_categories: Union[dict, list]):
        """
        [s0]

        Args:
            new_categories (list, dict):
                * list-like: all items must be unique and the number of items in the new categories must match the
                  existing number of categories.

                * dict-like: specifies a mapping from old categories to new. Categories not contained in the mapping are
                  passed through and extra categories in the mapping are ignored.
        Returns:

        """
        idx = self.df.index.rename_categories(new_categories)
        self.df.index = idx
        self.update_faults_index_rename()

        return self

    @_setdoc_pro([pn.CategoricalIndex.reorder_categories.__doc__, pn.CategoricalIndex.sort_values.__doc__])
    def reorder_series(self, new_categories: Union[list, np.ndarray]):
        """[s0] [s1]

        Args:
            new_categories (list): list with all series names in the desired order.

        Returns:
            Series
        """
        idx = self.df.index.reorder_categories(new_categories).sort_values()
        self.df = self.df.reindex(idx, copy=False)
        self.reset_order_series()
        self.update_faults_index_reorder()

        return self

    def modify_order_series(self, new_value: int, series_name: str):
        """
        Replace to the new location the old series

        Args:
            new_value (int): New location
            series_name (str): name of the series to be moved

        Returns:
            Series
        """
        group = self.df['order_series']
        assert np.isin(new_value, group), 'new_value must exist already in the order_surfaces group.'
        old_value = group[series_name]
        self.df['order_series'] = group.replace([new_value, old_value], [old_value, new_value])
        self.sort_series()
        self.update_faults_index_reorder()

        return self

    def sort_series(self):
        self.df.sort_values(by='order_series', inplace=True)
        self.df.index = self.df.index.reorder_categories(self.df.index.to_numpy())

    def update_faults_index_rename(self):
        idx = self.df.index
        self.faults.df.index = idx
        self.faults.faults_relations_df.index = idx
        self.faults.faults_relations_df.columns = idx

        #  This is a hack for qgrid
        #  We need to add the qgrid special columns to categories
        self.faults.faults_relations_df.columns = self.faults.faults_relations_df.columns.add_categories(
            ['index', 'qgrid_unfiltered_index'])

    def update_faults_index_reorder(self):
        idx = self.df.index
        self.faults.df = self.faults.df.reindex(idx, copy=False)
        self.faults.faults_relations_df = self.faults.faults_relations_df.reindex(idx, axis=0)
        self.faults.faults_relations_df = self.faults.faults_relations_df.reindex(idx, axis=1)

        self.faults.faults_relations_df.columns = self.faults.faults_relations_df.columns.add_categories(
            ['index', 'qgrid_unfiltered_index'])
        self.faults.set_default_faults_relations()


class MockFault:
    pass


class Stack(Series, Faults):
    """Class that encapsulates all type of geological features. So far is Series and
          Faults

         Args:
             features_names (Iterable): Names of the features
             fault_features (Iterable): List of features that are faults
             rel_matrix:

    """
    def __init__(self, features_names: Iterable = None, fault_features: Iterable = None,
                 rel_matrix: Iterable = None):

        if features_names is None:
            features_names = ['Default series']

        # Set unique df
        df_ = pn.DataFrame(np.array([[1, np.nan, False, False, False]]),
                               index=pn.CategoricalIndex(features_names, ordered=False),
                               columns=['order_series', 'BottomRelation', 'isActive', 'isFault', 'isFinite'])

        self.df = df_.astype({'order_series': int,
                              'BottomRelation': 'category',
                              'isActive': bool,
                              'isFault': bool,
                              'isFinite': bool})

        self.df['order_series'] = self.df['order_series'].astype(int)
        self.df['BottomRelation'] = pn.Categorical(['Erosion'], categories=['Erosion', 'Onlap', 'Fault'])

        self.faults = self
        self.faults_relations_df = pn.DataFrame(index=pn.CategoricalIndex(['Default series']),
                                                columns=pn.CategoricalIndex(['Default series', '']), dtype='bool')

        self.set_is_fault(series_fault=fault_features)
        self.set_fault_relation(rel_matrix=rel_matrix)
        self.n_faults = 0
        self._offset_faults = False














"""
Hecho
~~~~~

"""

# %% 
# These two lines are necessary only if gempy is not installed
import sys, os
import urllib

os.environ["aesara_FLAGS"] = "mode=FAST_RUN,device=cpu"
# Importing gempy
import gempy_legacy as gp

# Aux imports
import numpy as np
import pandas as pn
import matplotlib.pyplot as plt


# %%
# Loading surface points from repository:
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 
# With pandas we can do it directly from the web and with the right args
# we can directly tidy the data in gempy style:
# 

# %% 
dfs = []
data_amount = 'Full'

# First stratigraphic data
for letter in range(1, 10):

    try:
        dfs.append(pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master'
                               f'/Hecho/{data_amount}/H' +
                               str(letter) + '.csv', sep=';',
                               names=['X', 'Y', 'Z', 'surface', '_'], header=0))
    except urllib.error.HTTPError as e:
        print(e, letter)

# Also faults
for f in range(1, 4):
    fault_df = pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Hecho/F' +
                           str(f) + 'Line.csv',
                           sep=';', names=['X', 'Y', 'Z'], header=0)
    fault_df['surface'] = 'f' + str(f)
    dfs.append(fault_df)

# We put all the surfaces points together because is how gempy likes it:
surface_points = pn.concat(dfs, sort=True)
surface_points.reset_index(inplace=True, drop=False)
surface_points.tail()

# %%
# Now we do the same with the orientations:
# 

# %% 
orientations = pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Hecho/Sparse/Dips.csv',
                           sep=';',
                           names=['X', 'Y', 'Z', 'G_x', 'G_z', '_'], header=0)
# Orientation needs to belong to a surface. This is mainly to categorize to which series belong and to
# use the same color
orientations['surface'] = 0

# We fill the laking direction with a dummy value:
orientations['G_y'] = 0

# Drop unecesary data point
orientations.drop([1, 3, 4], inplace=True)
orientations

# %%
# Data initialization:
# ~~~~~~~~~~~~~~~~~~~~
# 
# Suggested size of the axis-aligned modeling box: Origin: 0 -0.5 0
# Maximum: 16 0.5 4.5
# 
# Suggested resolution: 0.05m (grid size 321 x 21 x 91)
# 

# %% 
geo_model = gp.create_model('Moureze')
geo_model = gp.init_data(geo_model, extent=[0, 16, -0.5, 0.5, 0, 4.5], resolution=[321, 21, 91],
                         surface_points_df=surface_points, orientations_df=orientations,
                         surface_name='surface',
                         add_basement=True)

# %% 
geo_model.orientations.df.at[5, 'surface']

# %% 
geo_model.orientations.df

# %%
# We need an orientation per series/fault. The faults does not have
# orientation so the easiest is to create an orientation from the surface
# points availablle:
# 

# %% 
f_names = ['f1', 'f2', 'f3']
for fn in f_names:
    fault_idx = geo_model.surface_points.df.index[geo_model.surface_points.df['surface'] == fn]
    gp.set_orientation_from_surface_points(geo_model, fault_idx)

# %%
# Now we can see how the data looks so far:
# 

# %% 
gp.plot_2d(geo_model)

# %%
# By default all surfaces belong to one unique series.
# 

# %% 
geo_model.surfaces

# %% 
geo_model.orientations.df.dtypes

# %%
# We will need to separate with surface belong to each series:
# 

# %% 
gp.map_stack_to_surfaces(geo_model, {'Fault1': 'f1', 'Fault2': 'f2', 'Fault3': 'f3'})

# %%
# However if we want the faults to offset the “Default series”, they will
# need to be more recent (higher on the pile). We can modify the order by:
# 

# %% 
geo_model.modify_order_series(4, 'Default series')

# %%
# Lastly, so far we did not specify which series/faults are actula faults:
# 

# %% 
geo_model.set_is_fault(['Fault1', 'Fault2', 'Fault3'])

# %%
# Now we are good to go:
# 

# %% 
gp.set_interpolator(geo_model, aesara_optimizer='fast_run', dtype='float64')

# %%
# The default range is always the diagonal of the extent. Since in this
# model data is very close we will need to reduce the range to 5-10% of
# that value:
# 

# %%
new_range = geo_model.get_additional_data().loc[('Kriging', 'range'), 'values'] * 0.2
geo_model.modify_kriging_parameters('range', new_range)


# %% 
gp.compute_model(geo_model, sort_surfaces=True, compute_mesh=False)

# %%
# Time
# ~~~~
# 
# -  GTX 2080 164 ms ± 2.03 ms per loop (mean ± std. dev. of 7 runs, 1
#    loop each)
# 

# %% 
gp.plot_2d(geo_model, cell_number=[10], series_n=3, show_scalar=True)

# %% 
gp.plot_2d(geo_model, cell_number=[10], show_data=True)

# %%
# sphinx_gallery_thumbnail_number = 3
gp.plot_3d(geo_model, kwargs_plot_structured_grid={'opacity': 8})


# %%
# Export data:
# ~~~~~~~~~~~~
# 
# The solution is stored in a numpy array of the following shape. Axis 0
# are the scalar fields of each correspondent series/faults in the
# following order (except basement):
# 

# %% 
geo_model.series

# %%
# For the surfaces, there are two numpy arrays, one with vertices and the
# other with triangles. Axis 0 is each surface in the order:
# 

# %% 
geo_model.surfaces

# %% 
np.save('Hecho_scalar', geo_model.solutions.scalar_field_matrix)


# %%
def write_property_to_gocad_voxet(propertyfilename, propertyvalues):
    """
    This function writes a numpy array into the right format for a gocad
    voxet property file. This assumet there is a property already added to the .vo file,
    and is just updating the file.
    propertyfile - string giving the path to the file to write
    propertyvalues - numpy array nz,ny,nx ordering and in float format
    """
    propertyvalues = propertyvalues.astype('>f4')  # big endian
    #     array = propertyvalues.newbyteorder()
    propertyvalues.tofile(propertyfilename)


# %%
write_property_to_gocad_voxet(f'hecho_sf_gempy_{data_amount}',
                              geo_model.solutions.scalar_field_matrix[3].reshape([321, 21, 91]).ravel('F'))

# %%
4



"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


@author: Fabian A. Stamm
"""

import numpy as np
import scipy.optimize as sop
from matplotlib import pyplot as plt

import copy
import matplotlib.pyplot as plt
from matplotlib import cm

import matplotlib.gridspec as gridspect

# Create cmap
from matplotlib.colors import ListedColormap
import matplotlib.colors as colors
import matplotlib.cm as cmx

import numpy as np
import pandas as pn
import scipy.stats as stats
import seaborn as sns

def loss_abs(estimate_s, true_s, u=1,o=1,u_f=1,o_f=1, r=1):
    """Absolute-error loss function.

        Args:
            estimate_s (int or float): Value estimate.
            true_s (int, float or np.array): True value.
            u (int or float, optional): Underestimation re-weighting factor.
            o (int or float, optional): Overestimation re-weighting factor.
            u_f (int or float, optional): Fatal underestimation re-weighting factor.
            o_f (int or float, optional): Fatal overestimation re-weighting factor.
            r (int or float, optional): Risk-affinity re-weighting factor.
        Returns:
            Loss incurred for an estimate given a true value.
            Based on absolute-error loss, i.e. the absolute distance of the estimate
            from the true value.

            true_s can be either a single value or an array of possible true values,
            the output will be a single determined absolute loss value or an array of determined
            loss values, accordingly.

            estimate_s has to be one single value.

    """
    true_s = np.array(true_s).astype(float)
    loss_s = np.zeros_like(true_s)
    underest = (estimate_s < true_s)
    overest = (estimate_s > true_s)
    loss_s[underest] = (true_s[underest] - estimate_s)  * (u * (r ** -0.5))
    loss_s[overest] = (estimate_s - true_s[overest]) * (o * r)
    if u_f != 1:
        underest_fatal = (estimate_s <= 0) & (true_s > 0)
        loss_s[underest_fatal] = (true_s[underest_fatal] - estimate_s) * (u_f * (r ** -0.5))
    if o_f != 1:
        overest_fatal = (estimate_s > 0) & (true_s <= 0)
        loss_s[overest_fatal] = np.abs((true_s[overest_fatal] - estimate_s)) * (o_f * r)
    return loss_s

def loss_abs_given_values(estimate_s, true_s, u=1,o=1,u_f=1,o_f=1, r=1):
    """Absolute-error loss function for the exclusive use with a single
    given true value.

            Args:
                estimate_s (int or float): Value estimate.
                true_s (int, float): True value.
                u (int or float, optional): Underestimation re-weighting factor.
                o (int or float, optional): Overestimation re-weighting factor.
                u_f (int or float, optional): Fatal underestimation re-weighting factor.
                o_f (int or float, optional): Fatal overestimation re-weighting factor.
                r (int or float, optional): Risk-affinity re-weighting factor.
            Returns:
                Loss incurred for an estimate given one determined true value.
                Based on absolute-error loss, i.e. the absolute distance of the estimate
                from the true value.

        """
    if estimate_s < true_s:
        if estimate_s <= 0 and true_s > 0:
            loss_s = (true_s - estimate_s) * (u_f * (r ** -0.5))  # bad case of underestimation
        else:
            loss_s = (true_s - estimate_s) * (u * (r ** -0.5)) # normal underestimation
    elif estimate_s > true_s:
        if estimate_s > 0 and true_s <= 0:
            loss_s = (estimate_s - true_s) * (o_f)  # bad case of overestimation
        else:
            loss_s = (estimate_s - true_s) * (o * r)  # normal overestimation
    else:
        loss_s = 0
    return loss_s

def loss_sqr(estimate_s, true_s, u=1,o=1,u_f=1,o_f=1, r=1):
    """Squared-error loss function.

        Args:
            estimate_s (int or float): Value estimate.
            true_s (int, float or np.array): True value.
            u (int or float, optional): Underestimation re-weighting factor.
            o (int or float, optional): Overestimation re-weighting factor.
            u_f (int or float, optional): Fatal underestimation re-weighting factor.
            o_f (int or float, optional): Fatal overestimation re-weighting factor.
            r (int or float, optional): Risk-affinity re-weighting factor.
        Returns:
            Loss incurred for an estimate of the value given a true value.
            Based on squared-error loss, i.e. the absolute distance of the estimate
            from the true value squared.

            true_s can be either a single value or an array of possible true values,
            the output will be a single determined absolute loss value or an array of determined
            loss values, accordingly.

            estimate_s has to be one single value.

    """
    return np.square(loss_abs(estimate_s, true_s, u,o,u_f,o_f, r))

def loss_sqr_given_values(estimate_s, true_s, u=1,o=1,u_f=1,o_f=1, r=1):
    """Squared-error loss function for the exclusive use with a single
        given true value.

                Args:
                    estimate_s (int or float): Value estimate.
                    true_s (int, float): True value.
                    u (int or float, optional): Underestimation re-weighting factor.
                    o (int or float, optional): Overestimation re-weighting factor.
                    u_f (int or float, optional): Fatal underestimation re-weighting factor.
                    o_f (int or float, optional): Fatal overestimation re-weighting factor.
                    r (int or float, optional): Risk-affinity re-weighting factor.
                Returns:
                    Loss incurred for an estimate given one determined true value.
                    Based on squared-error loss, i.e. the absolute distance of the estimate
                    from the true value squared.
    """
    return loss_abs_given_values(estimate_s, true_s, u,o,u_f,o_f, r)**2

def expected_loss_for_estimate(estimate_s, true_s, function='absolute', u=1,o=1,u_f=1,o_f=1, r=1):
    """Function to attain expected loss for an estimate
    given a range of possible true values by taking the mean of all
    possible loss realizations.

        Args:
            estimate_s (int or float): Value estimate.
            true_s (int, float or np.array): True value.
            function ('absolute'(default) or 'squared'): Use of absolute-error or
                squared-error loss function.
            u (int or float, optional): Underestimation re-weighting factor.
            o (int or float, optional): Overestimation re-weighting factor.
            u_f (int or float, optional): Fatal underestimation re-weighting factor.
            o_f (int or float, optional): Fatal overestimation re-weighting factor.
            r (int or float, optional): Risk-affinity re-weighting factor.
        Returns:
            Expected loss for a single estimate given a range of possible true values.

            Note: If only one possible true value is passed, the expected loss equals
            the actually incurred loss.

    """

    if function == 'absolute':
        return loss_abs(estimate_s, true_s, u,o,u_f,o_f, r).mean()
    elif function == 'squared':
        return loss_sqr(estimate_s, true_s, u, o, u_f, o_f, r).mean()
    else:
        print('Error: Type of loss function not recognized. '
              'Use "absolute" or "squared".')

def expected_loss_for_range(estimate_range, true_s, function='absolute', u=1,o=1,u_f=1,o_f=1, r=1):
    """Function to attain expected loss and Bayes action based on an absolute-error or
           squared error-loss function for a defined range of estimates.

                Args:
                    estimate_range (np.array): Range of value estimates.
                    true_s (np.array): Array of possible true value occurrences (from a probability distribution)
                    u (int or float, optional): Underestimation re-weighting factor.
                    o (int or float, optional): Overestimation re-weighting factor.
                    u_f (int or float, optional): Fatal underestimation re-weighting factor.
                    o_f (int or float, optional): Fatal overestimation re-weighting factor.
                    r (int, float or np.array, optional): Risk-affinity re-weighting factor.
                Returns:
                    [0]: Expected loss for the range of estimates.
                    [1]: Bayes action (estimate with minimal expected loss)
                    [2]: Expected loss of Bayes action.
        """
    expected_loss_s = lambda estimate_s, r: expected_loss_for_estimate(estimate_s, true_s, function, u, o, u_f, o_f, r)
    loss_e = [expected_loss_s(e, r) for e in estimate_range]
    bayes_action = sop.fmin(expected_loss_s, -40, args=(r,), disp=False)
    bayes_action_loss_e = expected_loss_s(bayes_action, r)
    return loss_e, bayes_action, bayes_action_loss_e

def expected_loss_plot(estimate_range, true_s, risk_range=1, function='absolute', u=1,o=1,u_f=1,o_f=1,
                        verbose=False):
    """Function to plot expected losses and the Bayes action for a range of estimates
    relative to a distribution of possible true values.
    It is possible to plot this for several risk factors at once.

            Args:
                estimate_range (np.array): Range of value estimates.
                true_s (np.array): Array of possible true value occurrences (from a probability distribution)
                u (int or float, optional): Underestimation re-weighting factor.
                o (int or float, optional): Overestimation re-weighting factor.
                u_f (int or float, optional): Fatal underestimation re-weighting factor.
                o_f (int or float, optional): Fatal overestimation re-weighting factor.
                r (int, float or np.array, optional): Risk-affinity re-weighting factor.
            Returns:
                Plot of expected losses for risk neutrality
                or several risk factors.

    """
    ax = plt.subplot(111)
    if isinstance(risk_range, (int,float)):
        r_range=[risk_range]
    else:
        r_range=risk_range
    for r in r_range:
        loss_e, bayes_a, bayes_a_loss_e = expected_loss_for_range(estimate_range, true_s, function, u,o,u_f,o_f, r)
        _color = next(ax._get_lines.prop_cycler)
        plt.plot(estimate_range, loss_e, label="r =" + str(r), color=_color['color'])
        plt.scatter(bayes_a, bayes_a_loss_e, s=70,
                        color=_color['color'])  # , label = "Bayes action r "+str(r))
        plt.vlines(bayes_a, 0, 10 * np.max(loss_e), color=_color['color'], linestyles="--")
        if verbose == True:
            print("Bayes action (minimum) at risk r %.2f: %.2f --- expected loss: %.2f"\
                  % (r, bayes_a, bayes_a_loss_e))
    plt.legend(loc="upper left", scatterpoints=1, title="Legend")
    plt.xlabel("Estimate")
    plt.ylabel("Expected loss")
    plt.xlim(estimate_range[0], estimate_range[-1])
    plt.ylim(0, 1.1 * np.max(loss_e))
    plt.grid()
    plt.show()

def loss_for_estimate(estimate_s, true_s, function='absolute', u=1,o=1,u_f=1,o_f=1, r=1):
    """Function to attain actually incurred loss for an estimate
    given a single true value.

        Args:
            estimate_s (int or float): Value estimate.
            true_s (int or float): True value.
            function ('absolute'(default) or 'squared'): Use of absolute-error or
                squared-error loss function.
            u (int or float, optional): Underestimation re-weighting factor.
            o (int or float, optional): Overestimation re-weighting factor.
            u_f (int or float, optional): Fatal underestimation re-weighting factor.
            o_f (int or float, optional): Fatal overestimation re-weighting factor.
            r (int or float, optional): Risk-affinity re-weighting factor.
        Returns:
            Loss for a single estimate given a single determined true value.
    """
    if function == 'absolute':
        return loss_abs_given_values(estimate_s, true_s, u,o,u_f,o_f, r)
    elif function == 'squared':
        return loss_sqr_given_values(estimate_s, true_s, u, o, u_f, o_f, r)
    else:
        print('Error: Type of loss function not recognized. '
              'Use "absolute" or "squared".')

def loss_for_range(estimate_range, true_s, function='absolute', u=1,o=1,u_f=1,o_f=1, r=1):
    """Function to attain loss and Bayes action based on an absolute-error or
               squared error-loss function for a defined range of estimates
               given one single true value.

                    Args:
                        estimate_range (np.array): Range of value estimates.
                        true_value (int or float): Array of possible true value occurrences (from a probability distribution)
                        u (int or float, optional): Underestimation re-weighting factor.
                        o (int or float, optional): Overestimation re-weighting factor.
                        u_f (int or float, optional): Fatal underestimation re-weighting factor.
                        o_f (int or float, optional): Fatal overestimation re-weighting factor.
                        r (int, float or np.array, optional): Risk-affinity re-weighting factor.
                    Returns:
                        [0]: Loss for the range of estimates.
                        [1]: Bayes action (estimate with minimal expected loss)
                        [2]: Expected loss of Bayes action.

                        Note: Since the is only one possible true value,
                        the Bayes action will always equal this value.
            """
    incurred_loss = lambda estimate_s, r: loss_for_estimate(estimate_s, true_s, function, u, o, u_f, o_f, r)
    loss_i = [incurred_loss(e, r) for e in estimate_range]
    bayes_action = sop.fmin(incurred_loss, -40, args=(r,), disp=False)
    bayes_action_loss = incurred_loss(bayes_action, r)
    return loss_i, bayes_action, bayes_action_loss

def loss_plot(estimate_range, true_s, risk_range=1, function='absolute', u=1,o=1,u_f=1,o_f=1,
                        verbose=False):
    """Function to plot losses for a range of estimates
    relative to a single given true value.
    It is possible to plot this for several risk factors at once.

            Args:
                estimate_range (np.array): Range of value estimates.
                true_s (int or float): Array of possible true value occurrences (from a probability distribution)
                u (int or float, optional): Underestimation re-weighting factor.
                o (int or float, optional): Overestimation re-weighting factor.
                u_f (int or float, optional): Fatal underestimation re-weighting factor.
                o_f (int or float, optional): Fatal overestimation re-weighting factor.
                r (int, float or np.array, optional): Risk-affinity re-weighting factor.
            Returns:
                Plot of losses for risk neutrality
                or several risk factors given a single determined true value.

    """
    ax = plt.subplot(111)
    if isinstance(risk_range, (int,float)):
        r_range=[risk_range]
    else:
        r_range=risk_range
    for r in r_range:
        loss_i, bayes_a, bayes_a_loss = loss_for_range(estimate_range, true_s, function, u,o,u_f,o_f, r)
        _color = next(ax._get_lines.prop_cycler)
        plt.plot(estimate_range, loss_i, label="r =" + str(r), color=_color['color'])
        plt.scatter(bayes_a, bayes_a_loss, s=70,
                        color=_color['color'])  # , label = "Bayes action r "+str(r))
        plt.vlines(bayes_a, 0, 10 * np.max(loss_i), color=_color['color'], linestyles="--")
        if verbose == True:
            print("Bayes action (minimum) at risk r %.2f: %.2f --- expected loss: %.2f"\
                  % (r, bayes_a, bayes_a_loss))
    plt.legend(loc="upper left", scatterpoints=1, title="Legend")
    plt.xlabel("Estimate")
    plt.ylabel("Expected loss")
    plt.xlim(estimate_range[0], estimate_range[-1])
    plt.ylim(0, 1.1 * np.max(loss_i))
    plt.grid()
    plt.show()


def plot_multiple_loss():

    def plot_axis(xvals, nor_l, loss, subplot_spec, e, depth=False):
        grid = gridspect.GridSpecFromSubplotSpec(1, 1, subplot_spec=subplot_spec)
        ax = fig.add_subplot(grid[:, :])
        # ax0.yaxis.set_visible(False)
        axr = ax.twinx()

        labels = 'Thickness Score' if depth is True else None
        c = default_red if depth is True else default_blue

        ax.plot(xvals, nor_l, color=c, linewidth=.5, label=labels)
        ax.fill_between(xvals, nor_l, 0, color=c, alpha=.8)
        ax.set_ylabel('Likelihood')
        ax.set_ylim(0, .16)
        ax.set_xlim(-25+6, 6+25)

        axr.set_ylim(0, 40)
        axr.set_ylabel('Expected Loss')

        ax.spines['top'].set_color('none')
        ax.spines['left'].set_color('none')
        if e == 0:
            axr.plot(xvals, loss, linewidth=3, color='white')
            axr.plot(xvals, loss, linewidth=2, color='#496155', label='Loss')
            axr.legend(frameon=True, facecolor='white', framealpha=1)

        else:
            axr.plot(xvals, loss, linewidth=3, color='white')
            axr.plot(xvals, loss, linewidth=2, color='#496155', label='Risk Neutral')
        if e%2 == 0:

            axr.yaxis.set_visible(False)
           # axr.yaxis.label.set_visible(False)

        else:

            for tick in axr.get_yticklines():
                tick.set_visible(False)
           # axr.yaxis.label.set_visible(False)
            ax.yaxis.set_visible(False)
        ax.xaxis.set_visible(False)
        axr.xaxis.set_visible(False)
        if e>3:
            ax.xaxis.set_visible(True)
            axr.xaxis.set_visible(True)
            ax.set_xlabel('Score')
        return ax, axr

    def res_score_loss(true_s, estimate_s, ov=1.25, uv_b=1.5, ov_b=2, risk_s=1):

        underest = (estimate_s < true_s)
        # loss_s = np.zeros((estimate_s.shape[0], true_s.shape[0]))
        underest_bad = (estimate_s <= 0) & (true_s > 0)
        overest = (estimate_s > true_s)
        overest_bad = (estimate_s > 0) & (true_s <= 0)
        a = underest * abs(true_s - estimate_s) * (risk_s ** -0.5)
        b = underest_bad * abs(true_s - estimate_s) * (uv_b * (risk_s ** -0.5))
        c = overest * abs(estimate_s - true_s) * (ov * risk_s)
        d = overest_bad * abs(estimate_s - true_s) * (ov_b * risk_s)
        loss_s = (a + b + c + d).mean(axis=1)
        return loss_s, (a,b,c,d)

    def abs_loss(true_s, estimate_s, ov=1, uv=1, risk_s=1, uv_b=1):
        print('foooooo')
        underest = (estimate_s < true_s)
        # loss_s = np.zeros((estimate_s.shape[0], true_s.shape[0]))
        underest_bad = 1# (estimate_s <= 0) & (true_s > 0)
        overest = (estimate_s > true_s)
        overest_bad = (estimate_s > 0) & (true_s <= 0)
        b = 0
        d = 0
        a = underest * abs(true_s - estimate_s) * (uv * risk_s)
      #  b = underest_bad * (true_s - estimate_s) * (uv_b * (risk_s))
        c = overest * abs(estimate_s - true_s) * (ov * risk_s)
        #d = overest_bad * (estimate_s - true_s) * (ov_b * risk_s)
        loss_s = (a + b + c + d).mean(axis=1)
        return loss_s, (a,b,c,d)

    def compute_values(mu, sigma, loss_type, x_range=(-24+6, 6+24), risk=1, depth=False):
        samples_size = 100000

        if x_range is None:
            x_max = mu + 5 * sigma
            x_min = mu - 5 * sigma
        else:
            x_min, x_max = x_range
       # xvals = np.linspace(-23+6, 6+23, 100)

        xvals = np.linspace(x_min, x_max, 100)
        nor_l = stats.norm.pdf(xvals, loc=mu, scale=sigma)
        samples = np.random.normal(loc=mu, scale=sigma, size=samples_size)

        if depth is True:
            #xvals = np.linspace(0, 60, 100)
            # nor_d = stats.norm.pdf(xvals, loc=30, scale=5)
            samples_depth = np.random.normal(loc=30, scale=5, size=samples_size)
            depth_score = samples_depth

            # Transformation to cost
            d_cost = - ((depth_score-depth_score.min()) / 8) ** 2.6
            samples = samples + d_cost
        elif depth is False:
            pass

        else:
            samples = depth

        if loss_type == 'abs':
            loss, partial = abs_loss(samples, xvals.reshape(-1, 1), 1, 1, risk_s=risk)
            #return xvals, nor_l, loss, partial
        elif loss_type == 'custom':
            loss, partial = res_score_loss(samples, xvals.reshape(-1, 1), risk_s=risk)

        if depth is True:
            return xvals, nor_l, loss, partial, samples, d_cost
        else:
            return xvals, nor_l, loss, partial

    sns.set(style="white")  # , rc={"axes.facecolor": (0, 0, 0, 0)})
    # Discrete cmap
    pal_disc = sns.cubehelix_palette(10, rot=-.25, light=.7)
    pal_disc_l = sns.cubehelix_palette(10)
    my_cmap = ListedColormap(pal_disc)
    my_cmap_l = ListedColormap(pal_disc_l)

    # Continuous cmap
    pal_cont = sns.cubehelix_palette(250, rot=-.25, light=.7)
    pal_cont_l = sns.cubehelix_palette(250)

    my_cmap_full = ListedColormap(pal_cont)
    my_cmap_full_l = ListedColormap(pal_cont_l)

    default_red = '#DA8886'
    default_blue = pal_cont.as_hex()[4]
    default_l = pal_disc_l.as_hex()[4]

    # %matplotlib notebook
    figsize = (12, 12)
    fig, axes = plt.subplots(0, 0, figsize=figsize, constrained_layout=False)
    sns.despine(left=True, bottom=True)

    gs_0 = gridspect.GridSpec(3, 2, figure=fig, hspace=0.1, wspace=0.009)

    axis = []

    mu = [12]*6
    sigma = [2.3, 2.3 ,4,4, 4, 4]
    type = ['abs', 'custom', 'abs', 'custom', 'abs', 'custom']
    for e, ax in enumerate(gs_0):

        if e==0:
            x, n, l, partial = compute_values(mu[e], sigma[e], type[e])
            ax_, axr_ = plot_axis(x, n, l, ax, e)
            ax_.vlines(mu[e], 0, 40, linestyles='--', alpha=.5)
            ax_.vlines(0, 0, 40, linestyles='--', alpha=.5)
            axr_.hlines(l.min(), -25, x[l.argmin()],
                       linestyles='--', alpha=.5)

        if e==1:
            x, n, l, partial = compute_values(mu[e], sigma[e], type[e])
            ax_, axr_ = plot_axis(x, n, l, ax, e)
            ax_.vlines(mu[e], 0, 40, linestyles='--', alpha=.5)
            ax_.vlines(0, 0, 40, linestyles='--', alpha=.5)
            labels = ['Underes.', 'Crit. Underest.', 'Overest.', 'Crit. Overest.']
            for i in range(4):
                axr_.plot(x, partial[i].mean(axis=1), '--', linewidth=1, label=labels[i])

            axr_.legend(loc=2, frameon=True, facecolor='white', framealpha=1)

        if e==2 or e==3:
            x, n, l, partial = compute_values(mu[e], sigma[e], type[e])
            ax_, axr_ = plot_axis(x, n, l, ax, e)
            axr_.plot(x[l.argmin()], 0, 'o',  color='#496155', markersize=12)
            if e == 2:
                axr_.hlines(l.min(), -25, x[l.argmin()],
                            linestyles='--', alpha=.5)

            x, n, l, partial = compute_values(mu[e], sigma[e], type[e], risk=.2)
            axr_.plot(x, l, linewidth=3, color='white')
            axr_.plot(x, l, linewidth=2, label='Risk Friendly')
            axr_.plot(x[l.argmin()], 0, 'o',  color='b', markersize=12)

            x, n, l, partial = compute_values(mu[e], sigma[e], type[e], risk=5)
            axr_.plot(x, l, linewidth=3, color='white')
            axr_.plot(x, l, linewidth=2, label='Risk Averse')
            axr_.plot(x[l.argmin()], 0, 'o',  color='orange', markersize=12)

            ax_.vlines(mu[e], 0, 40, linestyles='--', alpha=.5)
            ax_.vlines(0, 0, 40, linestyles='--', alpha=.5)
            if e==2:
                axr_.legend(frameon=True, facecolor='white', framealpha=1)

        if e > 3:
            x, n, l, partial, ss, d_cost = compute_values(mu[e], sigma[e], type[e],
                                                          x_range=(-24+6, 6+24),
                                                          risk=1, depth=True)
            ax_, axr_ = plot_axis(x, n, l, ax, e, depth=True)
            axr_.plot(x[l.argmin()], 0, 'o',  color='#496155', markersize=12)

            x_max = ss.mean() + 6 * ss.std()
            x_min = ss.mean() - 6 * ss.std()
            print(x_max, x_min, 'foo')
            print(ss.mean(), d_cost.mean())
            x, n, l, partial = compute_values(mu[e], sigma[e], type[e],
                                                      x_range=(-24 + 6, 6 + 24),
                                                      risk=.2, depth=ss)
            #print(_1.mean(),  d_cost.mean())
            axr_.plot(x, l, linewidth=3, color='white')
            axr_.plot(x, l, linewidth=2, label='Risk Friendly')
            axr_.plot(x[l.argmin()], 0, 'o',  color='b', markersize=12)

            x, n, l, partial = compute_values(mu[e], sigma[e], type[e],
                                                      x_range=(-24 + 6, 6 + 24),
                                                      risk=5, depth=ss)
          #  print(_1.mean(),  d_cost.mean())
            axr_.plot(x, l, linewidth=3, color='white')
            axr_.plot(x, l, linewidth=2, label='Risk Adverse')
            axr_.plot(x[l.argmin()], 0, 'o',  color='orange', markersize=12)

            k = sns.kdeplot(d_cost, ax=ax_, shade=True, label='Depth Score', color=default_red,
                        alpha=.8,linewidth=.5)
            #k.legend_.set_frame_on(True)
            sns.kdeplot(ss, ax=ax_, label='Final Score', shade=True, color=default_blue,
                        alpha=1, linewidth=.5)
            k.legend_.set_frame_on(True)
            k.legend_.set_alpha(1)

            if e == 4:
                k.legend_.set_visible(False)
            # ax_.vlines(mu[e], 0, 40, linestyles='--', alpha=.5)
            ax_.vlines(np.median(ss), 0, 40, linestyles='--', alpha=.5)
            ax_.vlines(0, 0, 40, linestyles='--', alpha=.5)
          #  ax_.legend(loc=2)

        axis.append((ax_, axr_))
    return k



import gempy_legacy as gp
import pandas as pn
import numpy as np
import os
import pytest

import gempy_legacy.core.grid
import gempy_legacy.core.surfaces
import gempy_legacy.core.data_modules.geometric_data
import gempy_legacy.core.data_modules.orientations
import gempy_legacy.core.data_modules.stack
import gempy_legacy.core.data_modules.surface_points


@pytest.fixture(scope="module")
def test_read_surface_points():
    surface_points = gempy_legacy.core.data_modules.surface_points.SurfacePoints()
    surface_points.read_surface_points(os.pardir + "/input_data/FabLessPoints_Points.csv", inplace=True)

    # Test setting series
    series = gempy_legacy.core.data_modules.stack.Series(series_distribution={"fault": 'MainFault',
                                            "Rest"                                   : ('SecondaryReservoir', 'Seal3', 'Reservoir', 'Overlying'),
                                                                              })
    surface_points.map_data_from_series(series, 'id')
    return surface_points


@pytest.fixture
def test_model_init(scope="module"):
    # Create empty model
    model = gp.Project()
    return model

@pytest.fixture
def test_create_series():
    series = gp.create_series(series_distribution={"fault": 'MainFault',
                                                   "Rest": ('SecondaryReservoir', 'Seal3', 'Reservoir', 'Overlying'),
                                                  })
    return series

@pytest.fixture()
def test_create_faults(test_create_series):
    faults = gempy_legacy.core.data_modules.stack.Faults(test_create_series)
    return faults

@pytest.fixture()
def test_create_formations():
    formations = gempy_legacy.core.Surfaces.Surfaces(values_array=np.arange(1, 8).reshape(-1, 1),
                                                     properties_names=np.array(['density']))
   # formations.set_surfaces_names(['MainFault', 'SecondaryReservoir','Seal',
   #                                 'Reservoir', 'Overlying'])

    return formations

@pytest.fixture(scope='module')
def test_load_model():
    model = gp.load_model(os.pardir + "/input_data/test_solution.pickle")
    return model


class TestModel:
    def test_modify_inteterfaces(self, load_model):
        load_model.modify_surface_points([0, 10, 5], X=[90, 20, 30], Y=[30, 20, 40])
        load_model.modify_surface_points(0, X=[60])

        load_model.modify_surface_points([0, 1, 4], X=[60])


class Testsurface_points:
    def test_map_all_to_data(self, test_create_series, test_read_surface_points, test_create_formations,
                             test_create_faults):
        test_read_surface_points.map_data_from_series(test_create_series)
        test_create_formations.set_surfaces_names(['MainFault', 'SecondaryReservoir', 'Seal',
                                                    'Reservoir', 'Overlying'])

        test_read_surface_points.map_formations_to_data(test_create_formations)
        test_read_surface_points.map_formations_to_data(test_create_formations)

        test_read_surface_points.map_data_from_faults(test_create_faults)
        test_read_surface_points.update_annotations()
        print(test_read_surface_points)

    def test_map_formations_to_data(self, test_read_surface_points, test_create_formations):
        test_read_surface_points.map_formations_to_data(test_create_formations)


class TestOrientations:
    @pytest.fixture(scope='class')
    def test_read_orientations(self):
        orientations = gempy_legacy.core.data_modules.orientations.Orientations()
        orientations.read_orientations(os.pardir + "/input_data/FabLessPoints_Foliations.csv", inplace=True)
        return orientations

    def test_map_all_to_data(self, test_create_series, test_read_orientations, test_create_formations,
                             test_create_faults):
        test_read_orientations.map_data_from_series(test_create_series)
        test_create_formations.set_surfaces_names(['MainFault', 'SecondaryReservoir', 'Seal',
                                                    'Reservoir', 'Overlying'])

        test_read_orientations.map_formations_to_data(test_create_formations)
        test_read_orientations.map_formations_to_data(test_create_formations)

        test_read_orientations.map_data_from_faults(test_create_faults)
        test_read_orientations.update_annotations()
        print(test_read_orientations)


class TestGrid:
    def test_set_regular_grid(self):
        # Test creating an empty list
        grid = gempy_legacy.core.grid.Grid()
        print(grid.create_regular_grid_3d([0,2000, 0, 2000, -2000, 0], [50, 50, 50]))

        # Test set regular grid by hand
        grid.create_regular_grid([0, 2000, 0, 2000, -2000, 0], [50, 50, 50])

    def test_grid_init(self):
        # Or we can init one of the default grids since the beginning by passing
        # the correspondant attributes
        grid = gempy_legacy.core.grid.Grid('regular_grid', extent=[0, 2000, 0, 2000, -2000, 0],
                                           resolution=[50, 50, 50])

    def test_grid_front(self):
        gp.create_grid('regular_grid', extent=[0, 2000, 0, 2000, -2000, 0],
                       resolution=[50, 50, 50])


class TestSeries:

    def test_set_series(self, test_read_surface_points):
        series = gempy_legacy.core.data_modules.stack.Series()
        # We can pass a pandas categories_df
        series.set_series_categories(pn.DataFrame({"fault": ['test2'],
                                        "Rest": 'SecondaryReservoir'}))

        # We can even pass an interface object since sometimes (GeoModeller) we
        # can import the surface in the same table
        series.set_series_categories(test_read_surface_points)
        print(series)

        # Test init series
        series = gempy_legacy.core.data_modules.stack.Series(series_distribution={"fault": 'MainFault',
                                                "Rest"                                   : ('SecondaryReservoir', 'Seal3', 'Reservoir', 'Overlying'),
                                                                                  })
        return series

    @pytest.fixture
    def test_series_front(self, test_model_init):
        model = test_model_init

        # Assigning series to surface as well as their order (timewise)
        gp.set_series(model, {"Fault_Series": 'Main_Fault',
                                 "Strat_Series": ('Sandstone_2', 'Siltstone',
                                                  'Shale', 'Sandstone_1')},
                      order_series=["Fault_Series", 'Strat_Series'],
                      order_formations=['Main_Fault',
                                        'Sandstone_2', 'Siltstone',
                                        'Shale', 'Sandstone_1', 'basement'
                                        ], verbose=0)

        print(model._stack)
        return model

    def test_sequential_pile(self, test_series_front):
        gp.get_sequential_pile(test_series_front)


class TestFaults:
    def test_set_faults(self, test_create_series):
        faults = gempy_legacy.core.data_modules.stack.Faults(test_create_series)
        faults.set_is_fault(['Rest'])
        print(faults)

    def test_default_faults(self, test_create_series):
        faults = gempy_legacy.core.data_modules.stack.Faults(test_create_series)
        print(faults)

    def test_set_fault_relations(self, test_create_faults):
        test_create_faults.set_fault_relation(np.array([[0, 1],
                                                        [0, 0]]))

        print(test_create_faults.faults_relations_df)


class TestFormations:
    def test_create_formation(self, test_create_formations):
        print(test_create_formations)

    def test_map_formations_from_series(self, test_create_formations, test_create_series):
        test_create_formations.map_formations_from_series(test_create_series)
        print(test_create_formations)

    def test_map_formations_from_series2(self, test_create_series):
        formations = gempy_legacy.core.Surfaces.Surfaces()
        formations.map_formations_from_series(test_create_series)
        print(formations)

    def test_set_formation_names(self, test_create_formations):
        test_create_formations.set_surfaces_names(['MainFault', 'SecondaryReservoir', 'Seal',
                                'Reservoir', 'Overlying'])

        print(test_create_formations)
        formations = gempy_legacy.core.Surfaces.Surfaces(values_array=np.arange(1, 8).reshape(-1, 1),
                                                         properties_names=np.array(['density']))

        formations.set_surfaces_names(['MainFault', 'SecondaryReservoir', 'Seal',
                                'Reservoir', 'Overlying'])
        print(formations)

        formations = gempy_legacy.core.Surfaces.Surfaces(values_array=np.arange(1, 2).reshape(-1, 1),
                                                         properties_names=np.array(['density']))

        formations.set_surfaces_names(['MainFault', 'SecondaryReservoir', 'Seal',
                                        'Reservoir', 'Overlying'])

        print(formations)
    def test_add_formation(self, test_create_formations):
        test_create_formations.add_basement()
        print(test_create_formations)
        # Test that break
        try:
            test_create_formations.add_basement('basement2')
        except AssertionError:
            print('assertion captured')

    def test_set_formations_values(self, test_create_formations):
        test_create_formations.set_formations_values(np.random.rand(2,2))
        test_create_formations.set_formations_values(np.random.rand(5,2))
        test_create_formations.set_formations_values(np.random.rand(10,3))


class TestSolution:
    def test_representation(self, load_model):
        sol = load_model.solutions
        sol.set_values(np.random.rand(4, 2, 3), compute_mesh=False)
        print(sol)

    def test_get_surfaces(self, load_model):
        model = load_model
        print(model.solutions)
        print(model.solutions.compute_all_surfaces())
        print(gp.get_surfaces(model))


def test_export_vtk(load_model):
    model = load_model
    gp.plot.export_to_vtk(model, os.path.dirname(__file__)+'/vtk/expert_test')


"""
Module initialisation for GemPy
Created on 21/10/2016

@author: Miguel de la Varga
"""
import sys
import os

import warnings

try:
    import faulthandler
    faulthandler.enable()
except Exception as e:  # pragma: no cover
    warnings.warn('Unable to enable faulthandler:\n%s' % str(e))


PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

# =================== Core ===================
from .core.model import Project, ImplicitCoKriging
from .core.data import AdditionalData, Options, KrigingParameters
from .core.data_modules.stack import Faults, Series
from .core.structure import Structure

from .core.grid import Grid
from .core.surfaces import Surfaces
from .core.data_modules.scaling_system import ScalingSystem
from .core.data_modules.orientations import Orientations
from .core.data_modules.surface_points import SurfacePoints
from .core.solution import Solution

# =================== API ===================
from .gempy_api import *
from .api_modules.getters import *
from .api_modules.setters import *
from .api_modules.io import *

# =================== Addons ===================
from .addons.gempy_to_rexfile import geomodel_to_rex

# =================== Plotting ===================
import gempy_legacy.plot.plot_api as plot
from .plot.plot_api import plot_2d, plot_3d
from .plot import _plot as _plot

# Assert at least pyton 3.10
assert sys.version_info[0] >= 3 and sys.version_info[1] >= 10, "GemPy requires Python 3.10 or higher"

__version__ = '2.3.2'

if __name__ == '__main__':
    pass



from . import docstring as doc_args
import re

# region Auxiliary


def _setdoc(docstring, indent=True, position='end'):
    """Copy the doc of a function or method into the decorated function.

    Args:
        docstring: docstring of to be added
        indent:
        position:

    Returns:

    """
    if type(docstring) is list:
        try:
            docstring = '----'.join(docstring)
        except TypeError:
            raise TypeError(str(docstring))

    def decor(func):

        if func.__doc__ is None:

            func.__doc__ = docstring
        else:

            if indent is True:
                aux = docstring.replace('\n', '\n\n        ')
                stop = aux.find('\nExamples\n-----')
                func.__doc__ += ' (inserted) \n        ' + aux[:stop]
            else:
                stop = docstring.find('\nExamples\n-----')
                if position == 'end':

                    func.__doc__ += docstring[:stop]
                else:
                    func.__doc__ = docstring[:stop] + '\n'+ func.__doc__
        return func

    return decor


def _setdoc_pro(docstring=[]):
    """This takes a list and places where it finds [s+0]"""
    if type(docstring) is not list:
        docstring = [docstring]

    def decor(func):

        if func.__doc__ is None:
            raise AttributeError('Add """"""" to the docstring')
           # print(func.__doc__)
        else:
            # Loop for the docstrings we pass looking for numbers
            for e, i in enumerate(docstring):
                # Find the location on the target method to insert comment
                marker = '[s'+str(e)+']'
                loc_0 = func.__doc__.find(marker) + len(marker)
                if loc_0 == -1:
                    print('marker not found')
                break_loc = i.find('\n\n')
                if break_loc == 0:
                    break_loc = i[2:].find('\n\n') + 2

                text = i[:break_loc]
                if text == -1:
                    text = 'No break found'
                text = text.replace('\n    ', '')
                func.__doc__ = func.__doc__[:loc_0] + ' '+ text + func.__doc__[loc_0:]

        # Find all the arg_string markers
        loc_1 = 0
        text = []
        marker = '[s_'
        for e in range(10):
            #print(len(func.__doc__), func.__doc__)
            #print('loc_1', loc_1)
            #print(func.__doc__)
            loc_0 = func.__doc__[loc_1:].find(marker)
            if loc_0 == -1:
                break
            else:
                loc_0 +=  loc_1
            #print('Here it is: ', func.__doc__[loc_1+len(text):loc_1+len(text)+50])
            #print(loc_0)

            end_marker = func.__doc__[loc_0:].find(']')
            loc_1 = loc_0 + end_marker
            arg_string = func.__doc__[loc_0 + 3:loc_1]
            #print('arg_string ',arg_string)
            try:
                text = getattr(doc_args, arg_string)
            except AttributeError as e:
                 print(e, func)

            # This 2 is to add the type to the string
            loc_1 = loc_0 + end_marker
            func.__doc__ = func.__doc__[:loc_0 - 2] + ' ' + text + func.__doc__[loc_1+1:]

        return func

    return decor
# endregion



import warnings
from typing import Union

import numpy as np
import pandas as pn

from .geometric_data import GeometricData
from ..surfaces import Surfaces
from ...utils import docstring as ds
from ...utils.meta import _setdoc_pro


@_setdoc_pro([Surfaces.__doc__, ds.coord, ds.surface_sp])
class SurfacePoints(GeometricData):
    """
    Data child with specific methods to manipulate interface data. It is initialize without arguments to give
    flexibility to the origin of the data.

    Args:
        surfa ces (:class:`Surfaces`): [s0]
        coord (np.ndarray): [s1]
        surface (list[str]): [s2]


    Attributes:
          df (:class:`pn.core.frame.DataFrames`): Pandas data frame containing the necessary information respect
          the surface points of the model
    """

    def __init__(self, surfaces: Surfaces, coord=None, surface=None):

        super().__init__(surfaces)
        self._columns_i_all = ['X', 'Y', 'Z', 'surface', 'series', 'X_std', 'Y_std', 'Z_std',
                               'order_series', 'surface_number']

        self._columns_i_1 = ['X', 'Y', 'Z', 'X_c', 'Y_c', 'Z_c', 'surface', 'series', 'id',
                             'order_series', 'isFault', 'smooth']

        self._columns_rep = ['X', 'Y', 'Z', 'surface', 'series']
        self._columns_i_num = ['X', 'Y', 'Z', 'X_c', 'Y_c', 'Z_c']
        self._columns_rend = ['X', 'Y', 'Z', 'smooth', 'surface']
        self.set_surface_points(coord, surface)

        self.df['order_series'] = self.df['order_series'].astype('int')
        self.df['id'] = self.df['id'].astype('int')

    @_setdoc_pro([ds.coord, ds.surface_sp])
    def set_surface_points(self, coord: np.ndarray = None, surface: list = None):
        """
        Set coordinates and surface columns on the df.

        Args:
            coord (np.ndarray): [s0]
            surface (list[str]): [s1]

        Returns:
            :class:`SurfacePoints`
        """
        self.df = pn.DataFrame(columns=['X', 'Y', 'Z', 'X_c', 'Y_c', 'Z_c', 'surface'], dtype=float)

        if coord is not None and surface is not None:
            self.df[['X', 'Y', 'Z']] = pn.DataFrame(coord)
            self.df['surface'] = surface

        self.df['surface'] = self.df['surface'].astype('category', copy=True)
        self.df['surface'] = self.df['surface'].cat.set_categories(self.surfaces.df['surface'].values)

        # Choose types
        self.init_dependent_properties()

        # Add nugget columns
        self.df['smooth'] = 2e-6

        assert ~self.df['surface'].isna().any(), 'Some of the surface passed does not exist in the Formation' \
                                                 'object. %s' % self.df['surface'][self.df['surface'].isna()]

        return self

    @_setdoc_pro([ds.x, ds.y, ds.z, ds.surface_sp, ds.idx_sp])
    def add_surface_points(self, x: Union[float, np.ndarray], y: Union[float, np.ndarray], z: Union[float, np.ndarray],
                           surface: Union[list, np.ndarray], idx: Union[int, list, np.ndarray] = None):
        """
        Add surface points.

        Args:
            x (float, np.ndarray): [s0]
            y (float, np.ndarray): [s1]
            z (float, np.ndarray): [s2]
            surface (list[str]): [s3]
            idx (Optional[int, list[int]): [s4]

        Returns:
           :class:`gempy.core.data_modules.geometric_data.SurfacePoints`

        """
        max_idx = self.df.index.max()

        if idx is None:
            idx = max_idx
            if idx is np.nan:
                idx = 0
            else:
                idx += 1

        if max_idx is not np.nan:
            self.df.loc[idx] = self.df.loc[max_idx]

        coord_array = np.array([x, y, z])
        assert coord_array.ndim == 1, 'Adding an interface only works one by one.'

        try:
            if self.surfaces.df.groupby('isBasement').get_group(True)['surface'].isin(surface).any():
                warnings.warn('Surface Points for the basement will not be used. Maybe you are missing an extra'
                              'layer at the bottom of the pile.')

            self.df.loc[idx, ['X', 'Y', 'Z']] = coord_array.astype('float64')
            
            # Check if elements in surface are categories in self.df['surface'] and if not add them
            #if not self.df['surface'].cat.categories.isin(surface).all():

            self._add_surface_to_list_from_new_surface_points_or_orientations(idx, surface)

        except TypeError as error:  # ToDO test this
            self.del_surface_points(idx)
            raise TypeError('The surface passed does not exist in the pandas categories. This may imply that'
                  'does not exist in the surface object either.')

        self.df.loc[idx, ['smooth']] = 1e-6
        
        self.df['surface'] = self.df['surface'].astype('category', copy=True)
        self.df['surface'] = self.df['surface'].cat.set_categories(self.surfaces.df['surface'].values)

        self.df['series'] = self.df['series'].astype('category', copy=True)
        self.df['series'] = self.df['series'].cat.set_categories(self.surfaces.df['series'].cat.categories)

        self.map_data_from_surfaces(self.surfaces, 'series', idx=idx)
        self.map_data_from_surfaces(self.surfaces, 'id', idx=idx)
        self.map_data_from_series(self.surfaces.series, 'order_series', idx=idx)

        self.sort_table()
        return self, idx

    
    @_setdoc_pro([ds.idx_sp])
    def del_surface_points(self, idx: Union[int, list, np.ndarray]):
        """Delete surface points.

        Args:
            idx (int, list[int]): [s0]

        Returns:
            :class:`gempy.core.data_modules.geometric_data.SurfacePoints`

        """
        self.df.drop(idx, inplace=True)
        return self

    @_setdoc_pro([ds.idx_sp, ds.x, ds.y, ds.z, ds.surface_sp])
    def modify_surface_points(self, idx: Union[int, list, np.ndarray], **kwargs):
        """Allows modification of the x,y and/or z-coordinates of an interface at specified dataframe index.

         Args:
             idx (int, list, np.ndarray): [s0]
             **kwargs:
                * X: [s1]
                * Y: [s2]
                * Z: [s3]
                * surface: [s4]

         Returns:
            :class:`gempy.core.data_modules.geometric_data.SurfacePoints`

         """
        idx = np.array(idx, ndmin=1)
        try:
            surface_names = kwargs.pop('surface')
            self.df.loc[idx, ['surface']] = surface_names
            self.map_data_from_surfaces(self.surfaces, 'series', idx=idx)
            self.map_data_from_surfaces(self.surfaces, 'id', idx=idx)
            self.map_data_from_series(self.surfaces.series, 'order_series', idx=idx)
            self.sort_table()
        except KeyError:
            pass

        # keys = list(kwargs.keys())
    #    is_surface = np.isin('surface', keys).all()

        # Check idx exist in the df
        assert np.isin(np.atleast_1d(idx), self.df.index).all(), 'Indices must exist in the' \
                                                                 ' dataframe to be modified.'

        # Check the properties are valid
        assert np.isin(list(kwargs.keys()), ['X', 'Y', 'Z', 'surface', 'smooth']).all(),\
            'Properties must be one or more of the following: \'X\', \'Y\', \'Z\', ' '\'surface\''
        # stack properties values
        values = np.array(list(kwargs.values()))

        # If we pass multiple index we need to transpose the numpy array
        if type(idx) is list or type(idx) is np.ndarray:
            values = values.T

        # Selecting the properties passed to be modified
        if values.shape[0] == 1:
            values = np.repeat(values, idx.shape[0])

        self.df.loc[idx, list(kwargs.keys())] = values

        return self

    @_setdoc_pro([ds.file_path, ds.debug, ds.inplace])
    def read_surface_points(self, table_source, debug=False, inplace=False,
                            kwargs_pandas: dict = None, **kwargs, ):
        """
        Read tabular using pandas tools and if inplace set it properly to the surface points object.

        Parameters:
            table_source (str, path object, file-like object or direct pandas data frame): [s0]
            debug (bool): [s1]
            inplace (bool): [s2]
            kwargs_pandas: kwargs for the panda function :func:`pn.read_csv`
            **kwargs:
                * update_surfaces (bool): If True add to the linked `Surfaces` object unique surface names read on
                  the csv file
                * coord_x_name (str): Name of the header on the csv for this attribute, e.g for coord_x. Default X
                * coord_y_name (str): Name of the header on the csv for this attribute. Default Y.
                * coord_z_name (str): Name of the header on the csv for this attribute. Default Z.
                * surface_name (str): Name of the header on the csv for this attribute. Default formation

        Returns:

        See Also:
            :meth:`GeometricData.read_data`
        """
        # TODO read by default either formation or surface

        if 'sep' not in kwargs:
            kwargs['sep'] = ','

        coord_x_name = kwargs.get('coord_x_name', "X")
        coord_y_name = kwargs.get('coord_y_name', "Y")
        coord_z_name = kwargs.get('coord_z_name', "Z")
        surface_name = kwargs.get('surface_name', "formation")

        if kwargs_pandas is None:
            kwargs_pandas = {}

        if 'sep' not in kwargs_pandas:
            kwargs_pandas['sep'] = ','

        if isinstance(table_source, pn.DataFrame):
            table = table_source
        else:
            table = pn.read_csv(table_source, **kwargs_pandas)

        if 'update_surfaces' in kwargs:
            if kwargs['update_surfaces'] is True:
                self.surfaces.add_surface(table[surface_name].unique())

        if debug is True:
            print('Debugging activated. Changes won\'t be saved.')
            return table
        else:
            assert {coord_x_name, coord_y_name, coord_z_name, surface_name}.issubset(table.columns), \
                "One or more columns do not match with the expected values " + str(table.columns)

            if inplace:
                c = np.array(self._columns_i_1)
                surface_points_read = table.assign(**dict.fromkeys(c[~np.in1d(c, table.columns)], np.nan))
                self.set_surface_points(surface_points_read[[coord_x_name, coord_y_name, coord_z_name]],
                                        surface=surface_points_read[surface_name])
            else:
                return table

    def set_default_surface_points(self):
        """
        Set a default point at the middle of the extent area to be able to start making the model
        """
        if self.df.shape[0] == 0:
            self.add_surface_points(0.00001, 0.00001, 0.00001, self.surfaces.df['surface'].iloc[0])
        return True

    def update_annotations(self):
        """
        Add a column in the Dataframes with latex names for each input_data paramenter.

        Returns:
            :class:`SurfacePoints`
        """
        point_num = self.df.groupby('id').cumcount()
        point_l = [r'${\bf{x}}_{\alpha \,{\bf{' + str(f) + '}},' + str(p) + '}$'
                   for p, f in zip(point_num, self.df['id'])]

        self.df['annotations'] = point_l
        return self



import warnings
from typing import Union, Iterable

import numpy as np
import pandas as pn
import sys

from .data import AdditionalData, Options, KrigingParameters
from .meta_data import MetaData
from .data_modules.stack import Stack, Faults, Series
from .interpolator import InterpolatorModel
from .solution import Solution
from ..plot.decorators import *
from ..utils.meta import _setdoc, _setdoc_pro
from ..utils.docstring import recompute_rf
from .surfaces import Surfaces
from .grid import Grid
from .data_modules.surface_points import SurfacePoints
from .data_modules.orientations import Orientations
from .data_modules.scaling_system import ScalingSystem


pn.options.mode.chained_assignment = None


class RestrictingWrapper(object):
    def __init__(self, w, accepted_members=['__repr__', '_repr_html_', '__str__']):
        self._w = w
        self._accepted_members = accepted_members

    def __repr__(self):
        return self._w.__repr__()

    def __getattr__(self, item):
        if item in self._accepted_members:
            return getattr(self._w, item)
        else:
            raise AttributeError(item)


@_setdoc_pro([Grid.__doc__, Faults.__doc__, Series.__doc__, Surfaces.__doc__, SurfacePoints.__doc__,Orientations.__doc__,
              ScalingSystem.__doc__, AdditionalData.__doc__, InterpolatorModel.__doc__, Solution.__doc__])
class ImplicitCoKriging(object):
    """This class handles all the mutation of the data objects of the model involved on the
     implicit cokriging ensuring the synchronization of all the members.

    Attributes:
        _grid (:class:`gempy.Grid`): [s0]
        _faults (:class:`gempy.Grid`): [s1]
        _stack (:class:`gempy.core.data_modules.stack.Stack`): [s2]
        _surfaces (:class:`gempy.Surfaces`): [s3]
        _surface_points (:class:`gempy.SurfacePoints`): [s4]
        _orientations (:class:`gempy.Orientations`): [s5]
        _rescaling (:class:`gempy.core.data_modules.geometric_data.Rescaling`): [s6]
        _additional_data (:class:`gempy.core.data.AdditionalData`): [s7]
        _interpolator (:class:`gempy.core.interpolator.InterpolatorModel`): [s8]
        solutions (:class:`gempy.core.solutions.Solutions`): [s9]


     """

    def __init__(self):

        self._grid = Grid()
        # Old way
        self._faults = Faults()
        self._stack = Series(self._faults)

        # New way
        self._stack = Stack()
        self._faults = self._stack.faults

        self._series = self._stack
        self._surfaces = Surfaces(self._stack)
        self._surface_points = SurfacePoints(self._surfaces)
        self._orientations = Orientations(self._surfaces)

        self._rescaling = ScalingSystem(self._surface_points, self._orientations,
                                        self._grid)
        self._additional_data = AdditionalData(self._surface_points,
                                               self._orientations, self._grid,
                                               self._faults,
                                               self._surfaces, self._rescaling)

        self._interpolator = InterpolatorModel(self._surface_points,
                                               self._orientations, self._grid,
                                               self._surfaces,
                                               self._stack, self._faults,
                                               self._additional_data)

        self.solutions = Solution(self._grid, self._surfaces, self._stack)

        # Previous values of sfai.
        self._sfai_order_0 = None

    @_setdoc_pro(Grid.__doc__)
    @property
    def grid(self):
        """ :class:`gempy.core.data.Grid` [s0]

        """
        return RestrictingWrapper(
            self._grid,
            accepted_members=['__repr__', '__str__', 'values', 'regular_grid',
                              'sections', 'centered_grid'])

    @_setdoc_pro(Faults.__doc__)
    @property
    def faults(self):
        """:class:`gempy.core.data_modules.stack.Faults` [s0]"""
        return RestrictingWrapper(self._faults,
                                  accepted_members=['__repr__', '_repr_html_',
                                                    'faults_relations_df'])

    @_setdoc_pro(Stack.__doc__)
    @property
    def stack(self):
        """:class:`gempy.core.data_modules.stack.Stack` [s0]"""
        return RestrictingWrapper(self._stack,
                                  accepted_members=['__repr__', '_repr_html_', 'df'])

    @_setdoc_pro(Series.__doc__)
    @property
    def series(self):
        """"""
        # warnings.warn(DeprecationWarning, 'series will be deprecated in the future.'
        #                                  'Use stack instead.')
        return RestrictingWrapper(self._stack,
                                  accepted_members=['__repr__', '_repr_html_', 'df'])

    @_setdoc_pro(Surfaces.__doc__)
    @property
    def surfaces(self):
        """:class:`gempy.core.data.Surfaces` [s0]"""
        return RestrictingWrapper(self._surfaces,
                                  accepted_members=['__repr__', '_repr_html_',
                                                    'colors', 'df'])

    @_setdoc_pro(SurfacePoints.__doc__)
    @property
    def surface_points(self):
        """:class:`gempy.core.data_modules.geometric_data.SurfacePoints` [s0]"""
        return RestrictingWrapper(self._surface_points,
                                  accepted_members=['__repr__', '_repr_html_', 'df'])

    @_setdoc_pro(Orientations.__doc__)
    @property
    def orientations(self):
        """:class:`gempy.core.data_modules.geometric_data.Orientations` [s0]"""
        return RestrictingWrapper(self._orientations,
                                  accepted_members=['__repr__', '_repr_html_', 'df'])

    @_setdoc_pro(ScalingSystem.__doc__)
    @property
    def rescaling(self):
        """:class:`gempy.core.data_modules.geometric_data.Rescaling` [s0]"""
        return RestrictingWrapper(self._rescaling)

    @_setdoc_pro(AdditionalData.__doc__)
    @property
    def additional_data(self):
        """:class:`gempy.core.data.AdditionalData` [s0]"""
        return RestrictingWrapper(self._additional_data,
                                  accepted_members=['__repr__', '_repr_html_',
                                                    'structure_data', 'options',
                                                    'kriging_parameters',
                                                    'kriging_data',
                                                    'rescaling_data'])

    @_setdoc_pro(InterpolatorModel.__doc__)
    @property
    def interpolator(self):
        """:class:`gempy.core.interpolator.InterpolatorModel` [s0]"""
        return RestrictingWrapper(self._interpolator,
                                  accepted_members=['__repr__', '_repr_html_',
                                                    'aesara_graph'])

    def _add_valid_idx_s(self, idx):
        if idx is None:
            idx = self._surface_points.df.index.max()
            if idx is np.nan:
                idx = 0
            else:
                idx += 1
        else:
            assert isinstance(idx, (
                int, list, np.ndarray)), 'idx must be an int or a list of ints'

        return idx

    def _add_valid_idx_o(self, idx):
        if idx is None:
            idx = self._orientations.df.index.max()
            if idx is np.nan:
                idx = 0
            else:
                idx += 1
        else:
            assert isinstance(idx, (
                int, list, np.ndarray)), 'idx must be an int or a list of ints'

        return idx

    @_setdoc_pro([AdditionalData.update_structure.__doc__,
                  InterpolatorModel.set_aesara_shared_structure.__doc__,
                  InterpolatorModel.modify_results_matrices_pro.__doc__,
                  InterpolatorModel.modify_results_weights.__doc__])
    def update_structure(self, update_aesara=None, update_series_is_active=True,
                         update_surface_is_active=True):
        """Update python and aesara structure parameters.

        [s0]
        [s1]

        Args:
            update_aesara (str{'matrices', 'weights'}):

                * matrices [s2]

                * weights [s3]

        Returns:
            :class:`gempy.core.data.Structure`
        """

        self._additional_data.update_structure()

        if update_series_is_active is True:
            len_series_i = self._additional_data.structure_data.df.loc[
                               'values', 'len series surface_points'] - \
                           self._additional_data.structure_data.df.loc[
                               'values', 'number surfaces per series']

            len_series_o = self._additional_data.structure_data.df.loc[
                'values', 'len series orientations'].astype(
                'int32')

            # Remove series without data
            non_zero_i = len_series_i.nonzero()[0]
            non_zero_o = len_series_o.nonzero()[0]
            non_zero = np.intersect1d(non_zero_i, non_zero_o)

            bool_vec = np.zeros_like(self._stack.df['isActive'], dtype=bool)
            bool_vec[non_zero] = True
            self._stack.df['isActive'] = bool_vec

        if update_surface_is_active is True:
            act_series = self._surfaces.df['series'].map(self._stack.df['isActive']).astype(bool)
            unique_surf_points = np.unique(self._surface_points.df['id'])
            if len(unique_surf_points) != 0:
                bool_surf_points = np.zeros_like(act_series, dtype=bool)
                bool_surf_points[unique_surf_points.astype('int') - 1] = True

                # This is necessary to find the intersection between orientations
                # (series) and  surface points
                self._surfaces.df['isActive'] = (
                                                        act_series & bool_surf_points) | self._surfaces.df['isBasement']
                self._surfaces.df['hasData'] = (
                        act_series | bool_surf_points)

        if update_aesara == 'matrices':
            self._interpolator.modify_results_matrices_pro()
        elif update_aesara == 'weights':
            self._interpolator.modify_results_weights()

        self._interpolator.set_aesara_shared_structure()
        return self._additional_data.structure_data

    # region Grid
    def update_from_grid(self):
        """Update objects dependent from the grid.

        """
        self._rescaling.rescale_data()
        self._interpolator.set_initial_results_matrices()

        if 'gravity' in self._interpolator.aesara_graph.output or 'magnetics' in self._interpolator.aesara_graph.output:
            self._interpolator.set_aesara_shared_l0_l1()

        # Check if grid is shared
        if hasattr(self._interpolator.aesara_graph.grid_val_T, 'get_value'):
            self._interpolator.aesara_graph.grid_val_T.set_value(
                self._grid.values_r.astype(self._interpolator.dtype))

    def set_active_grid(self, grid_name: Union[str, Iterable[str]], reset=False):
        """Set active a given or several grids.

        Args:
            grid_name (str, list[str]): Name of the grid you want to activate. Options are
             {regular, custom, topography, centered}
            reset (bool): If True set inactive all grids not in grid_name

        Returns:
            :class:`gempy.core.data.Grid`

        """
        if reset is True:
            self._grid.deactivate_all_grids()
        self._grid.set_active(grid_name)
        self.update_from_grid()
        print(f'Active grids: {self._grid.grid_types[self._grid.active_grids]}')

        return self._grid

    def get_active_grids(self):
        return self._grid.grid_types[self._grid.active_grids]

    def set_grid_object(self, grid: Grid, update_model=True):
        """Not implemented

        # TODO this should go to the api and let call all different grid types

        Args:
            grid:
            update_model:

        Returns:

        """
        raise NotImplementedError

    # @_setdoc(Grid.create_regular_grid.__doc__)
    @_setdoc_pro()
    def set_regular_grid(self, extent, resolution):
        """Set a regular grid, rescale data and initialize aesara solutions.

        Args:
            extent: [s_extent]
            resolution: [s_resolution]

        Returns:
            :class:`gempy.core.data.Grid`

        See Also:
            :class:`gempy.core.data.Grid.create_regular_grid`

            :class:`gempy.core.data.grid_modules.grid_types.RegularGrid`

        """
        if self._grid.regular_grid is None:
            self._grid.create_regular_grid(extent=extent, resolution=resolution)
        else:
            self._grid.regular_grid.set_regular_grid(extent=extent,
                                                     resolution=resolution)
            self._grid.set_active('regular')

        if self._grid.topography is not None and self._grid.topography.values.shape[
            0] != 0:
            self._grid.regular_grid.set_topography_mask(self._grid.topography)

        self.update_from_grid()
        print(f'Active grids: {self._grid.grid_types[self._grid.active_grids]}')
        return self._grid

    @_setdoc_pro()
    def set_custom_grid(self, custom_grid):
        """Set custom grid, rescale gird and initialize aesara solutions. foo

        Args:
            custom_grid: [s_coord]

        Returns:
            :class:`gempy.core.data.Grid`

        See Also:
            :class:`gempy.core.data.Grid.create_custom_grid`

            :class:`gempy.core.data.grid_modules.grid_types.CustomGrid`

        """
        if self._grid.custom_grid is None:
            self._grid.create_custom_grid(custom_grid)
        else:
            self._grid.custom_grid.set_custom_grid(custom_grid)
            self._grid.update_grid_values()

        self.update_from_grid()
        print(f'Active grids: {self._grid.grid_types[self._grid.active_grids]}')
        return self._grid

    @plot_set_topography
    def set_topography(self, source='random', set_mask=True, **kwargs):
        """Create a topography grid and activate it.

        Args:
            source:
                * 'gdal': Load topography from a raster file.
                * 'random': Generate random topography (based on a fractal grid).
                * 'saved': Load topography that was saved with the topography.save() function.
                  this is useful after loading and saving a heavy raster file with gdal once
                  or after saving a random topography with the save() function. This .npy file can then be set as
                  topography.

        Keyword Args:
            source = 'gdal':

                * filepath: path to raster file, e.g. '.tif', (for all file formats see
                  https://gdal.org/drivers/raster/index.html)

            source = 'random':

                * fd:         fractal dimension, defaults to 2.0

                * d_z:        maximum height difference. If none, last 20% of the model in z direction

                * extent:     extent in xy direction. If none, geo_model.grid.extent

                * resolution: desired resolution of the topography array. If none, geo_model.grid.resoution

            source = 'saved':

                * filepath:   path to the .npy file that was created using the topography.save() function

            source = 'numpy':
                * array: numpy array containing the data

        Returns:
             :class:`gempy.core.data.Grid`

        See Also:
            :class:`gempy.core.grid_modules.grid_types.Topography`

        """

        self._grid.create_topography(source, **kwargs)
        if set_mask is True:
            try:
                self._grid.regular_grid.set_topography_mask(self._grid.topography)
            except AttributeError:
                pass

        self.update_from_grid()
        print(f'Active grids: {self._grid.grid_types[self._grid.active_grids]}')
        return self._grid

    @_setdoc_pro(Grid.create_centered_grid.__doc__)
    def set_centered_grid(self, centers, radius, resolution=None):
        """[s0]

        Args:
            centers (numpy.ndarray[float, 3]): Location of the center of each kernel.
            radius (float): Distance from each center to create each XYZ point
            resolution (numpy.ndarray[3]): Number of voxels in each direction per kernel

        Returns:
            :class:`gempy.core.data.Grid`

        See Also:
            :class:`gempy.core.grid_modules.grid_types.CenteredGrid`

        """
        if self._grid.centered_grid is None:
            self._grid.create_centered_grid(centers, radius, resolution=resolution)
        else:
            self._grid.centered_grid.set_centered_grid(centers=centers,
                                                       radius=radius,
                                                       resolution=resolution)
            self._grid.update_grid_values()
        self.set_active_grid('centered')
        self.update_from_grid()
        # print(f'Active grids: {self._grid.grid_types[self._grid.active_grids]}')
        return self._grid

    @_setdoc_pro(Grid.create_section_grid.__doc__)
    def set_section_grid(self, section_dict):
        """[s0]

        Args:
            section_dict: [s_section_dict]

        Returns:
            :class:`gempy.core.grid_modules.grid_types.Sections`

        See Also:
            :class:`gempy.core.grid_modules.grid_types.Sections`
        """
        # TODO being able to change the regular grid associated to the section grid
        if self._grid.sections is None:
            self._grid.create_section_grid(section_dict=section_dict)
        else:
            self._grid.sections.set_sections(section_dict,
                                             regular_grid=self._grid.regular_grid)

        self.set_active_grid('sections')
        self.update_from_grid()
        return self._grid.sections

    # endregion

    # region Series
    def set_series_object(self):
        """
        Not implemented yet. Exchange the series object of the Model object.

        """
        raise NotImplementedError

    @_setdoc([Series.set_bottom_relation.__doc__], indent=False)
    def set_bottom_relation(self, series: Union[str, list],
                            bottom_relation: Union[str, list]):
        """"""
        self._stack.set_bottom_relation(series, bottom_relation)
        self._interpolator.set_aesara_shared_relations()
        return self._stack

    @_setdoc_pro(Stack.reset_order_series.__doc__)
    def add_features(self, features_list: Union[str, list], reset_order_series=True):
        """ Add series, update the categories dependent on them and reset the flow control.

        Args:
            features_list: (str, list): name or list of names of the series to apply the functionality
            reset_order_series: if true [s0]

        Returns:
            :class:`gempy.core.data_modules.stack.Stack`

        """
        self._stack.add_series(features_list, reset_order_series)
        self._surfaces.df['series'] = self._surfaces.df['series'].cat.add_categories(features_list)
        self._surface_points.df['series'] = self._surface_points.df['series'].cat.add_categories(features_list)
        self._orientations.df['series'] = self._orientations.df['series'].cat.add_categories(features_list)

        self.update_structure()
        self._interpolator.set_flow_control()
        self._interpolator.set_aesara_shared_kriging()
        return self._stack

    @_setdoc(Series.add_series.__doc__, indent=False)
    def add_series(self, series_list: Union[str, list], reset_order_series=True):
        warnings.warn('Series are getting renamed to Stack/features.'
                      'Please use add_features instead', DeprecationWarning, )
        return self.add_features(series_list, reset_order_series)

    @_setdoc_pro(Stack.reset_order_series.__doc__)
    def delete_features(self, indices: Union[str, list], reset_order_features=True,
                        remove_surfaces=False, remove_data=False):
        """ Delete series, update the categories dependent on them and reset the flow control.

        Args:
            indices (str, list): name or list of names of the series to apply the functionality
            reset_order_features: (bool): if true [s0]
            remove_surfaces (bool): if True remove the surfaces associated with the feature.
            remove_data (bool): if True remove the geometric data associated with the feature

        Returns:
             :class:`gempy.core.data_modules.stack.Stack`

        """
        indices = np.atleast_1d(indices)
        self._stack.delete_series(indices, reset_order_features)

        if remove_surfaces is True:
            for s in indices:
                self.delete_surfaces(
                    self._surfaces.df.groupby('series').get_group(s)['surface'],
                    remove_data=remove_data)

        self._surfaces.df['series'] = self._surfaces.df['series'].cat.remove_categories(indices)
        self._surface_points.df['series'] = self._surface_points.df['series'].cat.remove_categories(indices)
        self._orientations.df['series'] = self._orientations.df['series'].cat.remove_categories(indices)
        self.map_geometric_data_df(self._surface_points.df)
        self.map_geometric_data_df(self._orientations.df)

        self.update_structure()
        self._interpolator.set_aesara_shared_relations()
        self._interpolator.set_aesara_shared_kriging()
        self._interpolator.set_flow_control()
        return self._stack

    @_setdoc(delete_features.__doc__, indent=False)
    def delete_series(self, indices: Union[str, list], refactor_order_series=True,
                      remove_surfaces=False, remove_data=False):

        warnings.warn(DeprecationWarning,
                      'Series are getting renamed to Stack/features.'
                      'Please use delete_features instead')
        return self.delete_features(indices, refactor_order_series,
                                    remove_surfaces, remove_data)

    @_setdoc(Series.rename_series.__doc__, indent=False)
    def rename_features(self, new_categories: Union[dict, list]):
        """Rename features and update the category dependent on them.

        Args:
            new_categories (list, dict):
                * list-like: all items must be unique and the number of items in the new
                  categories must match the existing number of categories.

                * dict-like: specifies a mapping from old categories to new. Categories
                  not contained in the mapping are passed through and extra categories in the mapping are ignored.

        Returns:
            :class:`gempy.core.data_modules.stack.Stack`


        """
        self._stack.rename_series(new_categories)
        self._surfaces.df['series'] = self._surfaces.df['series'].cat.rename_categories(new_categories)
        self._surface_points.df['series'] = self._surface_points.df['series'].cat.rename_categories(new_categories)
        self._orientations.df['series'] = self._orientations.df['series'].cat.rename_categories(new_categories)
        return self._stack

    @_setdoc(rename_features.__doc__, indent=False)
    def rename_series(self, new_categories: Union[dict, list]):
        warnings.warn('Series are getting renamed to Stack/features.'
                      'Please use rename_features instead', DeprecationWarning)
        self.rename_features(new_categories)

    def modify_order_features(self, new_value: int, idx: str):
        """Modify order of the feature. Reorder categories of the link Surfaces, sort surface (reset the basement layer)
        remap the Stack and Surfaces to the corespondent dataframes, sort Geometric objects, update structure and
        reset the flow control objects.

        Args:
            new_value (int): New location
            idx (str): name of the feature to be moved

        Returns:
            :class:`gempy.core.data_modules.stack.Stack`

        """
        self._stack.modify_order_series(new_value, idx)

        self._surfaces.df['series'] = self._surfaces.df['series'].cat.reorder_categories(
            np.asarray(self._stack.df.index), ordered=False)

        self._surfaces.sort_surfaces()
        self._surfaces.set_basement()

        self.map_geometric_data_df(self._surface_points.df)
        self._surface_points.sort_table()
        self.map_geometric_data_df(self._orientations.df)
        self._orientations.sort_table()

        self._interpolator.set_flow_control()
        self.update_structure()
        return self._stack

    @_setdoc(Series.modify_order_series.__doc__, indent=False)
    def modify_order_series(self, new_value: int, idx: str):
        warnings.warn('Series are getting renamed to Stack/features.'
                      'Please use modify_order_features instead',
                      DeprecationWarning, )
        return self.modify_order_features(new_value, idx)

    def reorder_features(self, new_categories: Iterable[str]):
        """Reorder series. Reorder categories of the link Surfaces, sort surface (reset the basement layer)
        remap the Series and Surfaces to the corespondent dataframes, sort Geometric objects, update structure and
        reset the flow control objects.

        Args:
           new_categories (list): list with all series names in the desired order.

        Returns:
            :class:`gempy.core.data_modules.stack.Stack`
        """
        self._stack.reorder_series(new_categories)
        self._surfaces.df['series'] = self._surfaces.df['series'].cat.reorder_categories(
            np.asarray(self._stack.df.index),
            ordered=False)

        self._surfaces.sort_surfaces()
        self._surfaces.set_basement()

        self.map_geometric_data_df(self._surface_points.df)
        self._surface_points.sort_table()
        self.map_geometric_data_df(self._orientations.df)
        self._orientations.sort_table()

        self._interpolator.set_flow_control()
        self.update_structure(update_aesara='weights')
        return self._stack

    @_setdoc(reorder_features.__doc__, indent=False)
    def reorder_series(self, new_categories: Iterable[str]):
        warnings.warn('Series are getting renamed to Stack/features.'
                      'Please use reorder_features instead', DeprecationWarning)
        return self.reorder_features(new_categories)

    # endregion

    # region Faults
    def set_fault_object(self):
        """Not implemented"""
        raise NotImplementedError

    @_setdoc([Faults.set_is_fault.__doc__], indent=False)
    def set_is_fault(self, feature_fault: Union[str, list] = None,
                     toggle: bool = False,
                     change_color: bool = True, twofins=False):
        """
        Set a feature to fault and update all dependent objects of the Model.

        Args:

            feature_fault(str, list[str]): Name of the series which are faults
            toggle (bool): if True, passing a name which is already True will set it False.
            twofins (bool): If True, it allows to set several surfaces of a given geological feature to fault.
             This is behaviour is not tested and could have unexpected behaviour.
            change_color (bool): If True faults surfaces get the default fault color (light gray)

        Returns:
            :class:`gempy.core.data_modules.stack.Faults`

        See Also:
            :class:`gempy.core.data_modules.stack.Faults.set_is_fault`

        """
        feature_fault = np.atleast_1d(feature_fault)
        if twofins is False:
            for fault in feature_fault:
                if self._surfaces.df.shape[0] == 0:
                    aux_assert = True
                elif np.sum(self._surfaces.df.groupby('isBasement').get_group(False)[
                                'series'] == fault) < 2:
                    aux_assert = True
                else:
                    aux_assert = False

                assert aux_assert, \
                    'Having more than one fault in a series is generally rather bad. Better go' \
                    ' back to the function map_series_to_surfaces and give each fault its own' \
                    ' series. If you are really sure what you are doing, you can set twofins to' \
                    ' True to suppress this error.'

        self._faults.set_is_fault(feature_fault, toggle=toggle)
        if change_color:
            print(
                'Fault colors changed. If you do not like this behavior, set change_color to False.')
            self._surfaces.colors.make_faults_black(feature_fault)

        if toggle is True:
            already_fault = self._stack.df.loc[
                                feature_fault, 'BottomRelation'] == 'Fault'
            self._stack.df.loc[
                feature_fault[already_fault], 'BottomRelation'] = 'Erosion'
            print(feature_fault[already_fault])
            faults_list = list(self.surfaces.df[self.surfaces.df.series.isin(feature_fault[already_fault])]['surface'])
            for fault in faults_list:
                f_color = self._surfaces.df.index[self._surfaces.df['surface'] == fault][0]
                self._surfaces.colors.colordict[fault] = self._surfaces.colors._hexcolors_soft[f_color]
                self._surfaces.colors._set_colors()
            self._stack.df.loc[
                feature_fault[~already_fault], 'BottomRelation'] = 'Fault'
        else:
            self._stack.df.loc[feature_fault, 'BottomRelation'] = 'Fault'

        self._additional_data.structure_data.set_number_of_faults()
        self._interpolator.set_aesara_shared_relations()
        self._interpolator.set_aesara_shared_loop()
        if change_color:
            print(
                'Fault colors changed. If you do not like this behavior, set change_color to False.')
            self._surfaces.colors.make_faults_black(feature_fault)
        self.update_from_series(False, False, False)
        self.update_structure(update_aesara='matrices')
        return self._faults

    @_setdoc([Faults.set_is_finite_fault.__doc__], indent=False)
    def set_is_finite_fault(self, series_fault=None, toggle: bool = True):
        """"""
        s = self._faults.set_is_finite_fault(series_fault,
                                             toggle)  # change df in Fault obj
        # change shared aesara variable for infinite factor
        self._interpolator.set_aesara_shared_is_finite()
        return s

    @_setdoc([Faults.set_fault_relation.__doc__], indent=False)
    def set_fault_relation(self, rel_matrix):
        """"""
        self._faults.set_fault_relation(rel_matrix)

        # Updating
        self._interpolator.set_aesara_shared_fault_relation()
        self._interpolator.set_aesara_shared_weights()
        return self._faults.faults_relations_df

    # endregion

    # region Surfaces
    def set_surfaces_object(self):
        """
        Not implemented yet. Exchange the surface object of the Model object
        Returns:

        """
        raise NotImplementedError

    @_setdoc(Surfaces.add_surface.__doc__, indent=False)
    def add_surfaces(self, surface_list: Union[str, list], update_df=True):
        self._surfaces.add_surface(surface_list, update_df)
        self._surface_points.df['surface'] = self._surface_points.df['surface'].cat.add_categories(surface_list)
        self._orientations.df['surface'] = self._orientations.df['surface'].cat.add_categories(surface_list)
        self.update_structure()
        return self._surfaces

    @_setdoc_pro([Surfaces.update_id.__doc__])
    def delete_surfaces(self, indices: Union[str, Iterable[str]], update_id=True,
                        remove_data=True):
        """
        @TODO When implemeted activate geometric data, change remove data to False by default
        Delete a surface and update all related object.

        Args:
            indices (str, list): name or list of names of the series to apply the functionality
            update_id (bool): if true [s0]
            remove_data (bool): if true delete all GeometricData labeled with the given surface.

        Returns:
            :class:`gempy.core.data.Surfaces`

        """
        indices = np.atleast_1d(indices)
        self._surfaces.delete_surface(indices, update_id)

        if indices.dtype == int:
            surfaces_names = self._surfaces.df.loc[indices, 'surface']
        else:
            surfaces_names = indices

        if remove_data:
            self._surface_points.del_surface_points(
                self._surface_points.df[
                    self._surface_points.df.surface.isin(surfaces_names)].index)
            self._orientations.del_orientation(
                self._orientations.df[
                    self._orientations.df.surface.isin(surfaces_names)].index)

        self._surface_points.df['surface'] = self._surface_points.df['surface'].cat.remove_categories(surfaces_names)
        self._orientations.df['surface'] = self._orientations.df['surface'].cat.remove_categories(surfaces_names)
        self.map_geometric_data_df(self._surface_points.df)
        self.map_geometric_data_df(self._orientations.df)
        self._surfaces.colors.delete_colors(surfaces_names)

        if remove_data:
            self.update_structure(update_aesara='matrices')
            self.update_structure(update_aesara='weights')

        return self._surfaces

    @_setdoc(Surfaces.rename_surfaces.__doc__, indent=False)
    def rename_surfaces(self, to_replace: Union[dict], **kwargs):

        self._surfaces.rename_surfaces(to_replace, **kwargs)
        self._surface_points.df['surface'] = self._surface_points.df['surface'].cat.rename_categories(to_replace)
        self._orientations.df['surface'] = self._orientations.df['surface'].cat.rename_categories(to_replace)
        return self._surfaces

    @_setdoc(Surfaces.modify_order_surfaces.__doc__, indent=False)
    def modify_order_surfaces(self, new_value: int, idx: int,
                              series_name: str = None):
        """"""

        self._surfaces.modify_order_surfaces(new_value, idx, series_name)

        self.map_geometric_data_df(self._surface_points.df)
        self._surface_points.sort_table()
        self.map_geometric_data_df(self._orientations.df)
        self._orientations.sort_table()

        self.update_structure()
        return self._surfaces

    @_setdoc(Surfaces.add_surfaces_values.__doc__, indent=False)
    def add_surface_values(self, values_array: Iterable,
                           properties_names: Iterable[str] = np.empty(0)):
        self._surfaces.add_surfaces_values(values_array, properties_names)
        self.update_structure(update_aesara='matrices')
        return self._surfaces

    @_setdoc(Surfaces.delete_surface_values.__doc__, indent=False)
    def delete_surface_values(self, properties_names: list):
        self.delete_surface_values(properties_names)
        return self._surfaces

    @_setdoc(Surfaces.modify_surface_values.__doc__, indent=False)
    def modify_surface_values(self, idx, properties_names, values):
        self._surfaces.modify_surface_values(idx, properties_names, values)
        return self._surfaces

    @_setdoc(Surfaces.set_surfaces_values.__doc__, indent=False)
    def set_surface_values(self, values_array: Iterable,
                           properties_names: list = np.empty(0)):
        self._surfaces.set_surfaces_values(values_array, properties_names)
        return self._surfaces

    def map_stack_to_surfaces(self,
                              mapping_object: Union[dict, pn.Categorical] = None,
                              set_series=True, sort_geometric_data: bool = True,
                              remove_unused_series=True,
                              twofins=False
                              ):
        """Map series to surfaces and update all related objects accordingly to the following arguments:

        Args:
            mapping_object (dict, :class:`pandas.DataFrame`):

                * dict: keys are the series and values the surfaces belonging to that series

                * pandas.DataFrame: Dataframe with surfaces as index and a column series with the correspondent series
                  name of each surface

            set_series (bool): if True, if mapping object has non existing series they will be created.
            sort_geometric_data (bool): If true geometric data will be sorted accordingly to the new order of the
             series
            remove_unused_series (bool): if true, if an existing series is not assigned with a surface, it will get
             removed from the Series object.
            twofins (bool): If True, it allows to set several surfaces of a given geological feature to fault.
             This is behaviour is not tested and could have unexpected behaviour.

        Returns:
            :class:`gempy.core.data.Surfaces`

        """
        # Add New series to the series df
        if set_series is True:
            if type(mapping_object) is dict:
                series_list = list(mapping_object.keys())
                self._stack.add_series(series_list)
            elif isinstance(mapping_object, pn.Categorical):
                series_list = mapping_object['series'].values
                self._stack.add_series(series_list)
            else:
                raise AttributeError(
                    str(type(mapping_object)) + ' is not the right attribute type.')

        self._surfaces.map_series(mapping_object)

        # Here we remove the series that were not assigned to a surface
        if remove_unused_series is True:
            self._surfaces.df['series'] = self._surfaces.df['series'].cat.remove_unused_categories()
            unused_cat = self._stack.df.index[~self._stack.df.index.isin(
                self._surfaces.df['series'].cat.categories)]
            self._stack.delete_series(unused_cat)

        self._stack.reset_order_series()

        self.update_from_surfaces()
        self.update_from_series()

        if sort_geometric_data is True:
            self._surface_points.sort_table()
            self._orientations.sort_table()

        if set_series is True and self._stack.df.index.isin(['Basement']).any():
            aux = self._stack.df.index.drop('Basement').array
            self.reorder_features(np.append(aux, 'Basement'))

        if twofins is False:  # assert if every fault has its own series
            for serie in list(
                    self._faults.df[self._faults.df['isFault'] == True].index):
                assert np.sum(self._surfaces.df['series'] == serie) < 2, \
                    'Having more than one fault in a series is generally rather bad. Better give each ' \
                    'fault its own series. If you are really sure what you are doing, you can set ' \
                    'twofins to True to suppress this error.'

        self.update_structure()

        return self.surfaces

    @_setdoc(map_stack_to_surfaces.__doc__, indent=False)
    def map_series_to_surfaces(self, *args, **kwargs):
        return self.map_stack_to_surfaces(*args, **kwargs)

    # endregion

    # region Surface_points

    def set_surface_points_object(self, surface_points: SurfacePoints,
                                  update_model=True):
        """Not Implemented"""
        raise NotImplementedError

    @staticmethod
    def _check_possible_column_names(table, possible_candidates):
        possible_candidates = np.array(possible_candidates)
        return possible_candidates[np.isin(possible_candidates, table.columns)][0]

    @_setdoc_pro(SurfacePoints.set_surface_points.__doc__)
    def set_surface_points(self, table: pn.DataFrame, **kwargs):
        """Set coordinates and surface columns on the df.

        Args:
            table (pandas.Dataframe): table with surface points data.

        Keyword Args:
            bool add_basement: add a basement surface to the df.

        See Also:
            :class:`gempy.core.data_modules.geometric_data.SurfacePoints`

        """
        if isinstance(table,pn.DataFrame):
            table = table.reset_index(drop=True)
        else:
            raise ValueError('Input must be Pandas Dataframe')
        try:
            coord_x_name = kwargs.get('coord_x_name') if 'coord_x_name' in kwargs \
                else self._check_possible_column_names(table, ['X', 'x'])
            coord_y_name = kwargs.get('coord_y_name') if 'coord_y_name' in kwargs \
                else self._check_possible_column_names(table, ['Y', 'y'])
        except IndexError:
            raise ValueError('X and Y coordinates columns/values missing')

        try:
            coord_z_name = kwargs.get('coord_z_name') if 'coord_z_name' in kwargs \
                else self._check_possible_column_names(table, ['Z', 'z'])
        except IndexError:
            raise ValueError('Z coordinates column/values missing')

        surface_name = kwargs.get('surface_name') if 'surface_name' in kwargs \
            else self._check_possible_column_names(table,
                                                   ['surface', 'Surface', 'surfaces',
                                                    'surfaces', 'formations',
                                                    'formation'])
        update_surfaces = kwargs.get('update_surfaces', True)

        if update_surfaces is True:
            self.add_surfaces(table[surface_name].unique())

        c = np.array(self._surface_points._columns_i_1)
        surface_points_table = table.assign(
            **dict.fromkeys(c[~np.in1d(c, table.columns)], np.nan))
        self._surface_points.set_surface_points(
            surface_points_table[[coord_x_name, coord_y_name, coord_z_name]],
            surface=surface_points_table[surface_name])

        if 'add_basement' in kwargs:
            if kwargs['add_basement'] is True:
                self._surfaces.add_surface(['basement'])
                self.map_stack_to_surfaces({'Basement': 'basement'}, set_series=True)

        self.map_geometric_data_df(self._surface_points.df)
        self._rescaling.rescale_data()
        self.update_structure()
        return self._surface_points

    @_setdoc(Orientations.set_orientations.__doc__, indent=False, position='beg')
    def set_orientations(self, table: pn.DataFrame, **kwargs):
        """ Set coordinates, surface and orientation data.

        If both are passed pole vector has priority over orientation

        Args:
            table (pn.Dataframe): table with surface points data.

        Returns:
            :class:`gempy.core.data_modules.geometric_data.Orientations`

        See Also:
            :class:`gempy.core.data_modules.geometric_data.Orientations`

        """
        if isinstance(table,pn.DataFrame):
            table = table.reset_index(drop=True)
        else:
            raise ValueError('Input must be Pandas Dataframe')
        g_x_name = kwargs.get('G_x_name', 'G_x')
        g_y_name = kwargs.get('G_y_name', 'G_y')
        g_z_name = kwargs.get('G_z_name', 'G_z')
        azimuth_name = kwargs.get('azimuth_name', 'azimuth')
        dip_name = kwargs.get('dip_name', 'dip')
        polarity_name = kwargs.get('polarity_name', 'polarity')
        update_surfaces = kwargs.get('update_surfaces', False)

        coord_x_name = kwargs.get('coord_x_name') if 'coord_x_name' in kwargs \
            else self._check_possible_column_names(table, ['X', 'x'])
        coord_y_name = kwargs.get('coord_y_name') if 'coord_y_name' in kwargs \
            else self._check_possible_column_names(table, ['Y', 'y'])
        coord_z_name = kwargs.get('coord_z_name') if 'coord_z_name' in kwargs \
            else self._check_possible_column_names(table, ['Z', 'z'])
        surface_name = kwargs.get('surface_name') if 'surface_name' in kwargs \
            else self._check_possible_column_names(table,
                                                   ['surface', 'Surface', 'surfaces',
                                                    'surfaces', 'formations',
                                                    'formation', 'Formation'])

        if update_surfaces is True:
            self.add_surfaces(table[surface_name].unique())

        c = np.array(self._orientations._columns_o_1)
        orientations_read = table.assign(
            **dict.fromkeys(c[~np.in1d(c, table.columns)], np.nan))
        self._orientations.set_orientations(
            coord=orientations_read[[coord_x_name, coord_y_name, coord_z_name]],
            pole_vector=orientations_read[[g_x_name, g_y_name, g_z_name]].values,
            orientation=orientations_read[
                [azimuth_name, dip_name, polarity_name]].values,
            surface=orientations_read[surface_name])

        self.map_geometric_data_df(self._orientations.df)
        self._rescaling.rescale_data()
        self.update_structure()

        return self._orientations

    @_setdoc_pro(recompute_rf)
    @_setdoc(SurfacePoints.add_surface_points.__doc__, indent=False, position='beg')
    @plot_add_surface_points
    def add_surface_points(self, X, Y, Z, surface,
                           idx: Union[int, Iterable[int]] = None,
                           recompute_rescale_factor=False):
        """
        Args:
            X:
            Y:
            Z:
            surface (str):
            idx: Index of the point. If None, next available index will be used
            recompute_rescale_factor (bool): [s0].
        """

        surface = np.atleast_1d(surface)
        idx = self._add_valid_idx_s(idx)
        self._surface_points.add_surface_points(X, Y, Z, surface, idx)

        if recompute_rescale_factor is True or idx < 20:
            # This will rescale all data again
            self._rescaling.rescale_data()
            self._interpolator.set_aesara_shared_kriging()
        else:
            # This branch only recompute the added point
            self._rescaling.set_rescaled_surface_points(idx)
        self.update_structure(update_aesara='matrices')
        self._interpolator.set_aesara_shared_nuggets()

        return self._surface_points, idx

    @_setdoc(SurfacePoints.del_surface_points.__doc__, indent=False, position='beg')
    @plot_delete_surface_points
    def delete_surface_points(self, idx: Union[int, Iterable[int]]):
        self._surface_points.del_surface_points(idx)
        self.update_structure(update_aesara='matrices')
        return self._surface_points

    def delete_surface_points_basement(self):
        """Delete surface points belonging to the basement layer if any"""

        basement_name = \
            self._surfaces.df['surface'][self._surfaces.df['isBasement']].values[0]
        select = (self._surface_points.df['surface'] == basement_name)
        self.delete_surface_points(self._surface_points.df.index[select])
        return True

    @_setdoc_pro()
    @plot_move_surface_points
    def modify_surface_points(self, indices: Union[int, list],
                              recompute_rescale_factor=False, **kwargs):
        """Allows modification of the x,y and/or z-coordinates of an interface at specified dataframe index.

        Args:
            indices: [s_idx_sp]
            recompute_rescale_factor: [s_recompute_rf]

        Keyword Args:
                * X: [s_x]

                * Y: [s_y]

                * Z: [s_z]

                * surface: [s_surface_sp]

         Returns:
            :class:`gempy.core.data_modules.geometric_data.SurfacePoints`

         """

        keys = list(kwargs.keys())
        is_surface = np.isin('surface', keys).all()
        if is_surface:
            assert (
                ~self._surfaces.df[self._surfaces.df['isBasement']]['surface'].isin(
                    np.atleast_1d(kwargs['surface']))).any(), \
                'Surface points cannot belong to Basement. Add a new surface.'

        self._surface_points.modify_surface_points(indices, **kwargs)

        if recompute_rescale_factor is True or np.atleast_1d(indices).shape[0] < 20:
            # This will rescale all data again
            self._rescaling.rescale_data()
            self._interpolator.set_aesara_shared_kriging()
        else:
            # This branch only recompute the added point
            self._rescaling.set_rescaled_surface_points(indices)

        keys = list(kwargs.keys())
        is_surface = np.isin('surface', keys).all()
        if is_surface == True:
            self.update_structure(update_aesara='matrices')

        if 'smooth' in kwargs:
            self._interpolator.set_aesara_shared_nuggets()

        return self._surface_points

    # endregion

    # region Orientation
    def set_orientations_object(self, orientations: Orientations, update_model=True):
        raise NotImplementedError

    @_setdoc_pro()
    @plot_add_orientation
    def add_orientations(self, X, Y, Z, surface, pole_vector: Iterable = None,
                         orientation: Iterable = None, idx=None,
                         recompute_rescale_factor=False):
        """Add orientation.

        Args:
            X: [s_x]
            Y: [s_y]
            Z: [s_z]
            surface: [s_surface_sp]
            pole_vector: [s_pole_vector]
            orientation: [s_orientations]
            idx: [s_idx_sp]
            recompute_rescale_factor: [s_recompute_rf]

        Returns:
            :class:`gempy.core.data_modules.geometric_data.Orientations`

        """

        surface = list(np.atleast_1d(surface))
        idx = self._add_valid_idx_o(idx)

        self._orientations.add_orientation(X, Y, Z, surface, pole_vector=pole_vector,
                                           orientation=orientation, idx=idx)
        if recompute_rescale_factor is True or idx < 5:
            # This will rescale all data again
            self._rescaling.rescale_data()
        else:
            # This branch only recompute the added point
            self._rescaling.set_rescaled_orientations(idx)
        self.update_structure(update_aesara='weights')
        self._interpolator.set_aesara_shared_nuggets()

        return self._orientations, idx

    @_setdoc(Orientations.del_orientation.__doc__, indent=False, position='beg')
    @plot_delete_orientations
    def delete_orientations(self, idx: Union[list, int]):
        self._orientations.del_orientation(idx)
        self.update_structure(update_aesara='weights')
        return self._orientations

    @_setdoc(Orientations.modify_orientations.__doc__, indent=False, position='beg')
    @plot_move_orientations
    def modify_orientations(self, idx: list, **kwargs):

        idx = np.array(idx, ndmin=1)
        keys = list(kwargs.keys())
        is_surface = np.isin('surface', keys).all()
        self._orientations.modify_orientations(idx, **kwargs)
        self._rescaling.set_rescaled_orientations(idx)

        if is_surface:
            self.update_structure(update_aesara='weights')
        if 'smooth' in kwargs:
            self._interpolator.set_aesara_shared_nuggets()
        return self._orientations

    # endregion

    # region Options
    @_setdoc(Options.modify_options.__doc__, indent=False, position='beg')
    def modify_options(self, attribute, value):
        self._additional_data.options.modify_options(attribute, value)
        warnings.warn(
            'You need to recompile the aesara code to make it the changes in options.')

        return self._additional_data.options

    # endregion

    # region Kriging
    @_setdoc(KrigingParameters.modify_kriging_parameters.__doc__, indent=False,
             position='beg')
    def modify_kriging_parameters(self, attribute, value, **kwargs):
        self._additional_data.kriging_data.modify_kriging_parameters(attribute,
                                                                     value, **kwargs)
        self._interpolator.set_aesara_shared_kriging()
        if attribute == 'drift equations':
            self._interpolator.set_initial_results()
            self.update_structure()
        return self._additional_data.kriging_data

    # endregion

    # region rescaling
    @_setdoc(ScalingSystem.modify_rescaling_parameters.__doc__, indent=False,
             position='beg')
    def modify_rescaling_parameters(self, attribute, value):
        self._additional_data.rescaling_data.modify_rescaling_parameters(attribute,
                                                                         value)
        self._additional_data.rescaling_data.rescale_data()
        self._additional_data.update_default_kriging()

        return self._additional_data.rescaling_data

    # endregion

    # ======================================
    # --------------------------------------
    # ======================================

    def set_default_surface_point(self, **kwargs):
        """Set a default surface point if the df is empty. This is necessary for some
        type of functionality such as qgrid.

        Args:
            **kwargs: Same as :func:`gempy.core.data_modules.geometric_data.SurfacePoints.add_surface_points`

        Returns:
            :class:`gempy.core.data_modules.geometric_data.SurfacePoints`

        """
        if self._surface_points.df.shape[0] == 0:
            self.add_surface_points(0.00001, 0.00001, 0.00001,
                                    self._surfaces.df['surface'].iloc[0],
                                    recompute_rescale_factor=True, **kwargs)
        return self._surface_points

    def set_default_orientation(self, **kwargs):
        """Set a default orientation if the df is empty. This is necessary for some type of functionality such as qgrid

         Args:
             **kwargs: Same as :func::class:`gempy.core.data_modules.geometric_data.Orientations.add_orientation`

         Returns:
             :class:`gempy.core.data_modules.geometric_data.Orientations`

         """
        if self._orientations.df.shape[0] == 0:
            # TODO DEBUG: I am not sure that surfaces always has at least one entry. Check it
            self.add_orientations(.00001, .00001, .00001,
                                  self._surfaces.df['surface'].iloc[0],
                                  [0, 0, 1], recompute_rescale_factor=True, **kwargs)

    def set_default_surfaces(self):
        """Set two default surfaces if the df is empty. This is necessary for some type of functionality such as qgrid

         Returns:
             :class:`gempy.core.data.Surfaces`

         """
        if len(self._surfaces.df['surface']) != 0:
            self.delete_surfaces(self._surfaces.df['surface'])

        if self._surfaces.df.shape[0] == 0:
            self.add_surfaces(['surface1', 'surface2'])
        self.update_from_surfaces()
        return self._surfaces

    @_setdoc_pro()
    def set_extent(self, extent: Iterable):
        """
        Set project extent

        Args:
            extent: [s_extent]

        Returns:
            :class:`gempy.core.data.Grid`

        """
        extent = np.atleast_1d(extent)
        self._grid.extent = extent
        self._rescaling.set_rescaled_grid()
        return self._grid

    def update_from_series(self, reorder_series=True, sort_geometric_data=True,
                           update_interpolator=True):
        """Update all objects dependent on series.

        This method is a bit of a legacy and has been substituted by :meth:`rename_series` and :meth:`reorder_series`,
        however is useful if you want to make sure all objects are up to date with the latest changes on series.

        Args:

            reorder_series (bool): if True reorder all pandas categories accordingly to the series.df
            sort_geometric_data (bool): It True sort the geometric data after mapping the new order
            update_interpolator (bool): If True update the aesara shared variables dependent on the structure

        Returns:
            True
        """

        if reorder_series is True:
            self._surfaces.df['series'] = self._surfaces.df['series'].cat.reorder_categories(
                np.asarray(self._stack.df.index),
                ordered=False)
            self._stack.df.index = self._stack.df.index.reorder_categories(
                self._stack.df.index.array,
                ordered=False)
            self._surfaces.sort_surfaces()
            self.update_from_surfaces(set_categories_from_series=False,
                                      set_categories_from_surfaces=True,
                                      map_surface_points=False,
                                      map_orientations=False,
                                      update_structural_data=False)

        # Update surface is active from series does not work because you can have only a subset of surfaces of a
        # series active
        self._surfaces.df['isFault'] = self._surfaces.df['series'].map(
            self._faults.df['isFault'])
        self._surfaces.set_basement()

        # Add categories from series
        self._surface_points.set_series_categories_from_series(self._stack)
        self._orientations.set_series_categories_from_series(self._stack)

        self._surface_points.map_data_from_series(self._stack, 'order_series')
        self._orientations.map_data_from_series(self._stack, 'order_series')

        if sort_geometric_data is True:
            self._surface_points.sort_table()
            self._orientations.sort_table()

        self._additional_data.update_structure()
        # For the drift equations.
        self._additional_data.update_default_kriging()

        if update_interpolator is True:
            self._interpolator.set_aesara_shared_structure(reset_ctrl=True)

        return True

    def update_from_surfaces(self, set_categories_from_series=True,
                             set_categories_from_surfaces=True,
                             map_surface_points=True, map_orientations=True,
                             update_structural_data=True):
        """
        Update all objects dependt on surfaces.

        Args:
            set_categories_from_series (bool): If True update the pandas categories with the Series object
            set_categories_from_surfaces (bool): If True update the pandas categories with the surfaces object
            map_surface_points (bool): If True map the surface points fields with the Surfaces obejct
            map_orientations (bool): If True map the orientations fields with the Surfaces object
            update_structural_data (bool): If true update the Structure with the Surface object

        Returns:
            True
        """
        # Add categories from series
        if set_categories_from_series is True:
            self._surface_points.set_series_categories_from_series(
                self._surfaces.series)
            self._orientations.set_series_categories_from_series(
                self._surfaces.series)

        # Add categories from surfaces
        if set_categories_from_surfaces is True:
            self._surface_points.set_surface_categories_from_surfaces(self._surfaces)
            self._orientations.set_surface_categories_from_surfaces(self._surfaces)

        if map_surface_points is True:
            self._surface_points.map_data_from_surfaces(self._surfaces, 'series')
            self._surface_points.map_data_from_surfaces(self._surfaces, 'id')

        if map_orientations is True:
            self._orientations.map_data_from_surfaces(self._surfaces, 'series')
            self._orientations.map_data_from_surfaces(self._surfaces, 'id')

        if update_structural_data is True:
            self._additional_data.update_structure()

        return True

    # region aesara interface
    @_setdoc(InterpolatorModel.__doc__)
    def set_aesara_graph(self, interpolator: InterpolatorModel,
                         update_structure=True, update_kriging=True):
        """ Pass a aesara graph of a Interpolator instance other than the Model compose

        Use this method only if you know what are you doing!

        Args:
            interpolator (:class:`InterpolatorModel`): [s0]

        Returns:
             True """
        warnings.warn('This function is going to be deprecated. Use Model.set_aesara_function instead',
                      DeprecationWarning)
        self._interpolator.aesara_graph = interpolator.aesara_graph
        self._interpolator.aesara_function = interpolator.aesara_function
        self.update_additional_data(update_structure=update_structure,
                                    update_kriging=update_kriging)
        self.update_to_interpolator()
        return True

    # @_setdoc(InterpolatorModel.__doc__)
    def set_aesara_function(self, interpolator: InterpolatorModel,
                            update_structure=True, update_kriging=True):
        """Pass a aesara function and its correspondent graph from an Interpolator
         instance other than the Model compose

        Args:
            interpolator (:class:`gempy.core.interpolator.InterpolatorModel`): interpolator object
             with the compile graph.
            update_kriging (bool): if True update kriging parameters
            update_structure (bool): if Ture update structure

        Returns:
            bool: True

        See Also:
            :class:`gempy.core.interpolator.InterpolatorModel`
        """

        self._interpolator.aesara_graph = interpolator.aesara_graph
        self._interpolator.aesara_function = interpolator.aesara_function
        self.update_additional_data(update_structure=update_structure,
                                    update_kriging=update_kriging)
        self.update_to_interpolator()

        return True

    def update_additional_data(self, update_structure=True, update_kriging=True):
        if update_structure is True:
            self.update_structure(update_aesara='matrices')

        if update_kriging is True:
            print('Setting kriging parameters to their default values.')
            self._additional_data.update_default_kriging()

        return self._additional_data

    def update_to_interpolator(self, reset=True):
        """Update all shared parameters from the data objects

        Args:
            reset (bool): if True reset the flow control and initialize results arrays

        Returns:
            True
        """
        self._interpolator.set_all_shared_parameters()
        if reset is True:
            self._interpolator.reset_flow_control_initial_results()
        return True

    # endregion

    def map_geometric_data_df(self, d: pn.DataFrame):
        """
        Map a geometric data dataframe from the linked objects (at 07.2019 surfaces and series)

        Args:
            d (pn.DataFrame): Geometric data dataframe to be mapped

        Returns:
            DataFrame
        """
        d['series'] = d['surface'].map(
            self._surfaces.df.set_index('surface')['series'])
        d['id'] = d['surface'].map(
            self._surfaces.df.set_index('surface')['id']).astype(int)
        d['order_series'] = d['series'].map(self._stack.df['order_series']).astype(
            int)
        return d

    def set_surface_order_from_solution(self):
        """
        Order the surfaces respect the last computation. Therefore if you call this method,
        after sorting surface_points without recomputing you may get wrong results.

        Returns:
            Surfaces
        """

        field_at_surface_points = self.solutions.scalar_field_at_surface_points
        match type(field_at_surface_points):
            case np.ndarray:
                sfai_order = field_at_surface_points.sum(axis=0)
            case list:
                # concatenate
                sfai_order = np.concatenate(field_at_surface_points, axis=0)

                # Check if the order has changed
        if not np.array_equal(sfai_order, self._sfai_order_0):
            self._sfai_order_0 = sfai_order
            sel = self._surfaces.df['isActive'] & ~self._surfaces.df['isBasement']
            self._surfaces.df.loc[sel, 'sfai'] = sfai_order
            self._surfaces.df.sort_values(by=['series', 'sfai'], inplace=True, ascending=False)
            self._surfaces.reset_order_surfaces()
            self._surfaces.sort_surfaces()
            self._surfaces.set_basement()
            self._surface_points.df['id'] = self._surface_points.df['surface'].map(
                self._surfaces.df.set_index('surface')['id']).astype(int)
            self._orientations.df['id'] = self._orientations.df['surface'].map(
                self._surfaces.df.set_index('surface')['id']).astype(int)
            self._surface_points.sort_table()
            self._orientations.sort_table()
            self.update_structure()
        return self._surfaces


def Model(project_name='default_project'):
    """ Container class of all objects that constitute a GemPy model.

      In addition the class provides the methods that act in more than one of this class. Model is a child class of
      :class:`DataMutation` and :class:`MetaData`.

      """
    warnings.warn('This Class is going to be deprecated in GemPy 2.3. '
                  'Use Project instead.',
                  DeprecationWarning)
    return Project(project_name)


class Project(ImplicitCoKriging):
    """Container class of all objects that constitute a GemPy model.

    In addition the class provides all the methods you need to construct a geological
    model with :class:`ImplicitCoKriging`.

    See Also:
        :class:`MetaData`, :class:`ImplicitCoKriging`
    """

    def __init__(self, project_name='default_project'):

        self.meta = MetaData(project_name=project_name)
        super().__init__()

    def __repr__(self):
        return self.meta.project_name + ' ' + self.meta.date

    def new_model(self, name_project='default_project'):
        """Reset the model object."""
        self.__init__(name_project)

    def save_model_pickle(self, path=False):
        """
        Short term model storage. Object to a python pickle (serialization of python). Be aware that if the dependencies
        versions used to export and import the pickle differ it may give problems

        Args:
            path (str): path where save the pickle

        Returns:
            True
        """
        # Deleting qi attribute otherwise doesnt allow to pickle
        if hasattr(self, 'qi'):
            self.__delattr__('qi')

        sys.setrecursionlimit(10000)

        if not path:
            path = './' + self.meta.project_name
        import pickle
        with open(path + '.pickle', 'wb') as f:
            # Pickle the 'data' dictionary using the highest protocol available.
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
        return True

    @staticmethod
    def load_model_pickle(path):
        """
        Read InputData object from python pickle.

        Args:
           path (str): path where save the pickle

        Returns:
            :class:`gempy.core.model.Model`

        """
        import pickle
        with open(path, 'rb') as f:
            # The protocol version used is detected automatically, so we do not
            # have to specify it.
            model = pickle.load(f)
            return model

    def save_model(self, name=None, path=None, compress=True, save_solution=True):
        """
        Save model in new folder. Input data is saved as csv files. Solutions, extent and resolutions are saved as npy.

        Args:
            name (str): name of the newly created folder and the part of the files name
            path (str): path where save the model folder.
            save_solution (bool): if True, save the solution
            compress (bool): If true create a zip

        Returns:
            True
        """
        if name is None or path is None:
            from gempy_legacy.api_modules.io import default_path_and_name
            name, path = default_path_and_name(self, name, path)

        # save dataframes as csv
        self._surface_points.df.to_csv(f'{path}/{name}_surface_points.csv')
        self._surfaces.df.to_csv(f'{path}/{name}_surfaces.csv')
        self._orientations.df.to_csv(f'{path}/{name}_orientations.csv')
        self._stack.df.to_csv(f'{path}/{name}_series.csv')
        self._faults.df.to_csv(f'{path}/{name}_faults.csv')
        self._faults.faults_relations_df.to_csv(
            f'{path}/{name}_faults_relations.csv')
        self._additional_data.kriging_data.df.to_csv(
            f'{path}/{name}_kriging_data.csv')
        self._additional_data.rescaling_data.df.to_csv(
            f'{path}/{name}_rescaling_data.csv')
        self._additional_data.options.df.to_csv(f'{path}/{name}_options.csv')

        # # save resolution and extent as npy
        np.save(f'{path}/{name}_extent.npy', self._grid.regular_grid.extent)
        np.save(f'{path}/{name}_resolution.npy', self._grid.regular_grid.resolution)

        if self._grid.topography is not None:
            self._grid.topography.save(f'{path}/{name}_topography.npy')

        if self._grid.sections is not None:
            self._grid.sections.df.to_csv(f'{path}/{name}_sections.csv')
            np.save(f'{path}/{name}_sections.npy', self._grid.sections.values)

        if save_solution:
            self.save_solution(name, path)

        # if compress is True:
        #     shutil.make_archive(name, 'zip', path)
        #     shutil.rmtree(path)
        return True

    def save_solution(self, name=None, path=None):
        if self.solutions.lith_block is not None:
            np.save(f'{path}/{name}_lith_block.npy', self.solutions.lith_block)

    def read_data(self, source_i=None, source_o=None, add_basement=True, **kwargs):
        """
        Read data from a csv, or directly supplied dataframes

        Args:
            source_i: Path to the data bases of surface_points. Default os.getcwd(), or direct pandas data frame
            source_o: Path to the data bases of orientations. Default os.getcwd(), or direct pandas data frame
            add_basement (bool): if True add a basement surface. This wont be interpolated it just gives the values
            for the volume below the last surface.

        Keyword Args:
            update_surfaces (bool): True

        Returns:
            bool: True

        See Also:

            * :class:`gempy.core.data_modules.geometric_data.SurfacePoints.read_surface_points.`

            * :class:`gempy.core.data_modules.geometric_data.Orientations.read_orientations`
        """
        if 'update_surfaces' not in kwargs:
            kwargs['update_surfaces'] = True
        if 'path_i' in kwargs:
            source_i = kwargs['path_i']
        if 'path_o' in kwargs:
            source_o = kwargs['path_o']

        if isinstance(source_i, pn.DataFrame) or source_i:
            self._surface_points.read_surface_points(source_i, inplace=True,
                                                     **kwargs)
        if isinstance(source_o, pn.DataFrame) or source_o:
            self._orientations.read_orientations(source_o, inplace=True, **kwargs)
        if add_basement is True:
            self._surfaces.add_surface(['basement'])
            self.map_stack_to_surfaces({'Basement': 'basement'}, set_series=True)
        self._rescaling.rescale_data()

        self._additional_data.update_structure()
        self._additional_data.update_default_kriging()
        return True

    @_setdoc_pro()
    def get_data(self, itype='data', verbosity=0, numeric=False):
        """Method to return the data stored in :class:`panda.DataFrame` within a
        :class:`gempy.core.model.Project` data object.

        Args:
            itype: [s_itype]
            numeric(bool): Return only the numerical values of the dataframe. This is much lighter database for storing
                traces
            verbosity (int): Number of properties shown

        Returns:
            pandas.DataFrame: Data Object df.

        """

        if verbosity == 0:
            show_par_f = self._orientations._columns_rend
            show_par_i = self._surface_points._columns_rend
        elif verbosity == 1:
            show_par_f = self._orientations._columns_o_1
            show_par_i = self._surface_points._columns_i_1

        if numeric:
            show_par_f = self._orientations._columns_o_num
            show_par_i = self._surface_points._columns_i_num
            dtype = 'float'

        if itype == 'orientations':
            raw_data = self._orientations.df[show_par_f]  # .astype(dtype)
            # Be sure that the columns are in order when used for operations
            if numeric:
                raw_data = raw_data[
                    ['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', 'dip', 'azimuth',
                     'polarity']]
        elif itype == 'surface_points' or itype == 'surface points':
            raw_data = self._surface_points.df[show_par_i]  # .astype(dtype)
            # Be sure that the columns are in order when used for operations
            if numeric:
                raw_data = raw_data[
                    ['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', 'dip', 'azimuth',
                     'polarity']]
        elif itype == 'data':
            raw_data = pn.concat(
                [self._surface_points.df[show_par_i],  # .astype(dtype),
                 self._orientations.df[show_par_f]],  # .astype(dtype)],
                keys=['surface_points', 'orientations'],
                sort=False)
            # Be sure that the columns are in order when used for operations
            if numeric:
                raw_data = raw_data[
                    ['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', 'dip', 'azimuth',
                     'polarity']]

        elif itype == 'surfaces':
            raw_data = self._surfaces
        elif itype == 'series':
            raw_data = self._stack
        elif itype == 'faults':
            raw_data = self._faults
        elif itype == 'faults_relations_df' or itype == 'faults_relations':
            raw_data = self._faults.faults_relations_df
        elif itype == 'additional data' or itype == 'additional_data':
            raw_data = self._additional_data
        elif itype == 'kriging':
            raw_data = self._additional_data.kriging_data
        else:
            raise AttributeError(
                'itype has to be \'data\', \'additional data\', \'surface_points\', \'orientations\','
                ' \'surfaces\',\'series\', \'faults\' or \'faults_relations_df\'')

        return raw_data

    def get_additional_data(self):
        return self._additional_data.get_additional_data()



# These two lines are necessary only if GemPy is not installed
import sys, os
sys.path.append("../..")

# Importing GemPy
import gempy_legacy as gp
from gempy_legacy.assets.geophysics import GravityPreprocessing
from gempy_legacy.core.grid_modules.grid_types import CenteredGrid

# Importing auxiliary libraries
import numpy as np
import pandas as pn
import matplotlib.pyplot as plt
import pytest


def test_irregular_grid():
    g = CenteredGrid()
    g.set_centered_grid(np.array([0, 0, 0]), resolution=[5, 5, 5], radius = [100,100,100])
